import pandas as pd
import re
from .constants import PARSING_SCHEME, SEASON_PAGE_URL
from pyquery import PyQuery as pq
from ..decorators import float_property_decorator, int_property_decorator
from .. import utils
from .roster import Roster
from .schedule import Schedule


class Team(object):
    """
    An object containing all of a team's season information.

    Finds and parses all team stat information and identifiers, such as rank,
    name, and abbreviation, and sets them as properties which can be directly
    read from for easy reference.

    Parameters
    ----------
    team_data : string
        A string containing all of the rows of stats for a given team. If
        multiple tables are being referenced, this will be comprised of
        multiple rows in a single string.
    rank : int
        A team's position in the league based on the number of points they
        obtained during the season.
    year : string (optional)
        The requested year to pull stats from.
    """
    def __init__(self, team_data, rank, year=None):
        self._year = year
        self._rank = rank
        self._abbreviation = None
        self._name = None
        self._games_played = None
        self._minutes_played = None
        self._field_goals = None
        self._field_goal_attempts = None
        self._field_goal_percentage = None
        self._three_point_field_goals = None
        self._three_point_field_goal_attempts = None
        self._three_point_field_goal_percentage = None
        self._two_point_field_goals = None
        self._two_point_field_goal_attempts = None
        self._two_point_field_goal_percentage = None
        self._free_throws = None
        self._free_throw_attempts = None
        self._free_throw_percentage = None
        self._offensive_rebounds = None
        self._defensive_rebounds = None
        self._total_rebounds = None
        self._assists = None
        self._steals = None
        self._blocks = None
        self._turnovers = None
        self._personal_fouls = None
        self._points = None
        self._opp_field_goals = None
        self._opp_field_goal_attempts = None
        self._opp_field_goal_percentage = None
        self._opp_three_point_field_goals = None
        self._opp_three_point_field_goal_attempts = None
        self._opp_three_point_field_goal_percentage = None
        self._opp_two_point_field_goals = None
        self._opp_two_point_field_goal_attempts = None
        self._opp_two_point_field_goal_percentage = None
        self._opp_free_throws = None
        self._opp_free_throw_attempts = None
        self._opp_free_throw_percentage = None
        self._opp_offensive_rebounds = None
        self._opp_defensive_rebounds = None
        self._opp_total_rebounds = None
        self._opp_assists = None
        self._opp_steals = None
        self._opp_blocks = None
        self._opp_turnovers = None
        self._opp_personal_fouls = None
        self._opp_points = None

        self._parse_team_data(team_data)

    def _parse_team_data(self, team_data):
        """
        Parses a value for every attribute.

        This function looks through every attribute with the exception of
        '_rank' and retrieves the value according to the parsing scheme and
        index of the attribute from the passed HTML data. Once the value is
        retrieved, the attribute's value is updated with the returned result.

        Note that this method is called directly once Team is invoked and does
        not need to be called manually.

        Parameters
        ----------
        team_data : string
            A string containing all of the rows of stats for a given team. If
            multiple tables are being referenced, this will be comprised of
            multiple rows in a single string.
        """
        for field in self.__dict__:
            # The rank attribute is passed directly to the class during
            # instantiation.
            if field == '_rank' or \
               field == '_year':
                continue
            value = utils._parse_field(PARSING_SCHEME,
                                       team_data,
                                       str(field)[1:])
            setattr(self, field, value)

    @property
    def dataframe(self):
        """
        Returns a pandas DataFrame containing all other class properties and
        values. The index for the DataFrame is the string abbreviation of the
        team, such as 'DET'.
        """
        fields_to_include = {
            'abbreviation': self.abbreviation,
            'assists': self.assists,
            'blocks': self.blocks,
            'defensive_rebounds': self.defensive_rebounds,
            'field_goal_attempts': self.field_goal_attempts,
            'field_goal_percentage': self.field_goal_percentage,
            'field_goals': self.field_goals,
            'free_throw_attempts': self.free_throw_attempts,
            'free_throw_percentage': self.free_throw_percentage,
            'free_throws': self.free_throws,
            'games_played': self.games_played,
            'minutes_played': self.minutes_played,
            'name': self.name,
            'offensive_rebounds': self.offensive_rebounds,
            'opp_assists': self.opp_assists,
            'opp_blocks': self.opp_blocks,
            'opp_defensive_rebounds': self.opp_defensive_rebounds,
            'opp_field_goal_attempts': self.opp_field_goal_attempts,
            'opp_field_goal_percentage': self.opp_field_goal_percentage,
            'opp_field_goals': self.opp_field_goals,
            'opp_free_throw_attempts': self.opp_free_throw_attempts,
            'opp_free_throw_percentage': self.opp_free_throw_percentage,
            'opp_free_throws': self.opp_free_throws,
            'opp_offensive_rebounds': self.opp_offensive_rebounds,
            'opp_personal_fouls': self.opp_personal_fouls,
            'opp_points': self.opp_points,
            'opp_steals': self.opp_steals,
            'opp_three_point_field_goal_attempts':
            self.opp_three_point_field_goal_attempts,
            'opp_three_point_field_goal_percentage':
            self.opp_three_point_field_goal_percentage,
            'opp_three_point_field_goals': self.opp_three_point_field_goals,
            'opp_total_rebounds': self.opp_total_rebounds,
            'opp_turnovers': self.opp_turnovers,
            'opp_two_point_field_goal_attempts':
            self.opp_two_point_field_goal_attempts,
            'opp_two_point_field_goal_percentage':
            self.opp_two_point_field_goal_percentage,
            'opp_two_point_field_goals': self.opp_two_point_field_goals,
            'personal_fouls': self.personal_fouls,
            'points': self.points,
            'rank': self.rank,
            'steals': self.steals,
            'three_point_field_goal_attempts':
            self.three_point_field_goal_attempts,
            'three_point_field_goal_percentage':
            self.three_point_field_goal_percentage,
            'three_point_field_goals': self.three_point_field_goals,
            'total_rebounds': self.total_rebounds,
            'turnovers': self.turnovers,
            'two_point_field_goal_attempts':
            self.two_point_field_goal_attempts,
            'two_point_field_goal_percentage':
            self.two_point_field_goal_percentage,
            'two_point_field_goals': self.two_point_field_goals
        }
        return pd.DataFrame([fields_to_include], index=[self._abbreviation])

    @int_property_decorator
    def rank(self):
        """
        Returns an ``int`` of the team's rank based on the number of points
        they score per game.
        """
        return self._rank

    @property
    def abbreviation(self):
        """
        Returns a ``string`` of the team's abbreviation, such as 'DET' for the
        Detroit Pistons.
        """
        return self._abbreviation

    @property
    def schedule(self):
        """
        Returns an instance of the Schedule class containing the team's
        complete schedule for the season.
        """
        return Schedule(self._abbreviation, self._year)

    @property
    def roster(self):
        """
        Returns an instance of the Roster class containing all players for the
        team during the season with all career stats.
        """
        return Roster(self._abbreviation, self._year)

    @property
    def name(self):
        """
        Returns a ``string`` of the team's full name, such as 'Detroit
        Pistons'.
        """
        return self._name

    @int_property_decorator
    def games_played(self):
        """
        Returns an ``int`` of the total number of games the team has played
        during the season.
        """
        return self._games_played

    @int_property_decorator
    def minutes_played(self):
        """
        Returns an ``int`` of the total number of minutes played by all players
        on the team during the season.
        """
        return self._minutes_played

    @int_property_decorator
    def field_goals(self):
        """
        Returns an ``int`` of the total number of field goals the team has made
        during the season.
        """
        return self._field_goals

    @int_property_decorator
    def field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of field goals the team has
        attempted during the season.
        """
        return self._field_goal_attempts

    @float_property_decorator
    def field_goal_percentage(self):
        """
        Returns a ``float`` of the percentage of field goals made divided by
        the number of attempts. Percentage ranges from 0-1.
        """
        return self._field_goal_percentage

    @int_property_decorator
    def three_point_field_goals(self):
        """
        Returns an ``int`` of the total number of three point field goals the
        team has made during the season.
        """
        return self._three_point_field_goals

    @int_property_decorator
    def three_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of three point field goals the
        team has attempted during the season.
        """
        return self._three_point_field_goal_attempts

    @float_property_decorator
    def three_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the percentage of three point field goals made
        divided by the number of attempts. Percentage ranges from 0-1.
        """
        return self._three_point_field_goal_percentage

    @int_property_decorator
    def two_point_field_goals(self):
        """
        Returns an ``int`` of the total number of two point field goals the
        team has made during the season.
        """
        return self._two_point_field_goals

    @int_property_decorator
    def two_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of two point field goals the
        team has attempted during the season.
        """
        return self._two_point_field_goal_attempts

    @float_property_decorator
    def two_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the percentage of two point field goals made
        divided by the number of attempts. Percentage ranges from 0-1.
        """
        return self._two_point_field_goal_percentage

    @int_property_decorator
    def free_throws(self):
        """
        Returns an ``int`` of the total number of free throws made during the
        season.
        """
        return self._free_throws

    @int_property_decorator
    def free_throw_attempts(self):
        """
        Returns an ``int`` of the total number of free throw attempts during
        the season.
        """
        return self._free_throw_attempts

    @float_property_decorator
    def free_throw_percentage(self):
        """
        Returns a ``float`` of the percentage of free throws made divided by
        the attempts. Percentage ranges from 0-1.
        """
        return self._free_throw_percentage

    @int_property_decorator
    def offensive_rebounds(self):
        """
        Returns an ``int`` of the total number of offensive rebounds the team
        has grabbed.
        """
        return self._offensive_rebounds

    @int_property_decorator
    def defensive_rebounds(self):
        """
        Returns an ``int`` of the total number of defensive rebounds the team
        has grabbed.
        """
        return self._defensive_rebounds

    @int_property_decorator
    def total_rebounds(self):
        """
        Returns an ``int`` of the total number of rebounds the team has
        grabbed.
        """
        return self._total_rebounds

    @int_property_decorator
    def assists(self):
        """
        Returns an ``int`` of the total number of field goals that were
        assisted.
        """
        return self._assists

    @int_property_decorator
    def steals(self):
        """
        Returns an ``int`` of the total number of times the team stole the ball
        from the opponent.
        """
        return self._steals

    @int_property_decorator
    def blocks(self):
        """
        Returns an ``int`` of the total number of times the team blocked an
        opponent's shot.
        """
        return self._blocks

    @int_property_decorator
    def turnovers(self):
        """
        Returns an ``int`` of the total number of times the team has turned the
        ball over.
        """
        return self._turnovers

    @int_property_decorator
    def personal_fouls(self):
        """
        Returns an ``int`` of the total number of times the team has fouled an
        opponent.
        """
        return self._personal_fouls

    @int_property_decorator
    def points(self):
        """
        Returns an ``int`` of the total number of points the team has scored
        during the season.
        """
        return self._points

    @int_property_decorator
    def opp_field_goals(self):
        """
        Returns an ``int`` of the total number of field goals the opponents
        made during the season.
        """
        return self._opp_field_goals

    @int_property_decorator
    def opp_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of field goals the opponents
        attempted during the season.
        """
        return self._opp_field_goal_attempts

    @float_property_decorator
    def opp_field_goal_percentage(self):
        """
        Returns a ``float`` of the percentage of field goals made divided by
        the number of attempts by the opponent. Percentage ranges from 0-1.
        """
        return self._opp_field_goal_percentage

    @int_property_decorator
    def opp_three_point_field_goals(self):
        """
        Returns an ``int`` of the total number of three point field goals the
        opponent made during the season.
        """
        return self._opp_three_point_field_goals

    @int_property_decorator
    def opp_three_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of three point field goals the
        opponent attempted during the season.
        """
        return self._opp_three_point_field_goal_attempts

    @float_property_decorator
    def opp_three_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the percentage of three point field goals made
        divided by the number of attempts by the opponent. Percentage ranges
        from 0-1.
        """
        return self._opp_three_point_field_goal_percentage

    @int_property_decorator
    def opp_two_point_field_goals(self):
        """
        Returns an ``int`` of the total number of two point field goals the
        opponent made during the season.
        """
        return self._opp_two_point_field_goals

    @int_property_decorator
    def opp_two_point_field_goal_attempts(self):
        """
        Returns an ``int`` of the total number of two point field goals the
        opponent attempted during the season.
        """
        return self._opp_two_point_field_goal_attempts

    @float_property_decorator
    def opp_two_point_field_goal_percentage(self):
        """
        Returns a ``float`` of the percentage of two point field goals made
        divided by the number of attempts by the opponent. Percentage ranges
        from 0-1.
        """
        return self._opp_two_point_field_goal_percentage

    @int_property_decorator
    def opp_free_throws(self):
        """
        Returns an ``int`` of the total number of free throws made during the
        season by the opponent.
        """
        return self._opp_free_throws

    @int_property_decorator
    def opp_free_throw_attempts(self):
        """
        Returns an ``int`` of the total number of free throw attempts during
        the season by the opponent.
        """
        return self._opp_free_throw_attempts

    @float_property_decorator
    def opp_free_throw_percentage(self):
        """
        Returns a ``float`` of the percentage of free throws made divided by
        the attempts by the opponent. Percentage ranges from 0-1.
        """
        return self._opp_free_throw_percentage

    @int_property_decorator
    def opp_offensive_rebounds(self):
        """
        Returns an ``int`` of the total number of offensive rebounds the
        opponent grabbed.
        """
        return self._opp_offensive_rebounds

    @int_property_decorator
    def opp_defensive_rebounds(self):
        """
        Returns an ``int`` of the total number of defensive rebounds the
        opponent grabbed.
        """
        return self._opp_defensive_rebounds

    @int_property_decorator
    def opp_total_rebounds(self):
        """
        Returns an ``int`` of the total number of rebounds the opponent
        grabbed.
        """
        return self._opp_total_rebounds

    @int_property_decorator
    def opp_assists(self):
        """
        Returns an ``int`` of the total number of field goals that were
        assisted by the opponent.
        """
        return self._opp_assists

    @int_property_decorator
    def opp_steals(self):
        """
        Returns an ``int`` of the total number of times the opponent stole the
        ball from the team.
        """
        return self._opp_steals

    @int_property_decorator
    def opp_blocks(self):
        """
        Returns an ``int`` of the total number of times the opponent blocked
        the team's shot.
        """
        return self._opp_blocks

    @int_property_decorator
    def opp_turnovers(self):
        """
        Returns an ``int`` of the total number of times the opponent turned the
        ball over.
        """
        return self._opp_turnovers

    @int_property_decorator
    def opp_personal_fouls(self):
        """
        Returns an ``int`` of the total number of times the opponent fouled the
        team.
        """
        return self._opp_personal_fouls

    @int_property_decorator
    def opp_points(self):
        """
        Returns an ``int`` of the total number of points the team has been
        scored on during the season.
        """
        return self._opp_points


class Teams:
    """
    A list of all NBA teams and their stats in a given year.

    Finds and retrieves a list of all NBA teams from
    www.basketball-reference.com and creates a Team instance for every team
    that participated in the league in a given year. The Team class comprises
    a list of all major stats and a few identifiers for the requested season.

    Parameters
    ----------
    year : string (optional)
        The requested year to pull stats from.
    """
    def __init__(self, year=None):
        self._teams = []

        self._retrieve_all_teams(year)

    def __getitem__(self, abbreviation):
        """
        Return a specified team.

        Returns a team's instance in the Teams class as specified by the team's
        abbreviation.

        Parameters
        ----------
        abbreviation : string
            An NBA team's three letter abbreviation (ie. 'DET' for Detroit
            Pistons).

        Returns
        -------
        Team instance
            If the requested team can be found, its Team instance is returned.

        Raises
        ------
        ValueError
            If the requested team is not present within the Teams list.
        """
        for team in self._teams:
            if team.abbreviation.upper() == abbreviation.upper():
                return team
        raise ValueError('Team abbreviation %s not found' % abbreviation)

    def __call__(self, abbreviation):
        """
        Return a specified team.

        Returns a team's instance in the Teams class as specified by the team's
        abbreviation. This method is a wrapper for __getitem__.

        Parameters
        ----------
        abbreviation : string
            An NBA team's three letter abbreviation (ie. 'DET' for Detroit
            Pistons).

        Returns
        -------
        Team instance
            If the requested team can be found, its Team instance is returned.
        """
        return self.__getitem__(abbreviation)

    def __repr__(self):
        """Returns a ``list`` of all NBA teams for the given season."""
        return self._teams

    def __iter__(self):
        """Returns an iterator of all of the NBA teams for a given season."""
        return iter(self.__repr__())

    def __len__(self):
        """Returns the number of NBA teams for a given season."""
        return len(self.__repr__())

    def _add_stats_data(self, teams_list, team_data_dict):
        """
        Add a team's stats row to a dictionary.

        Pass table contents and a stats dictionary of all teams to accumulate
        all stats for each team in a single variable.

        Parameters
        ----------
        teams_list : generator
            A generator of all row items in a given table.
        team_data_dict : {str: {'data': str, 'rank': int}} dictionary
            A dictionary where every key is the team's abbreviation and every
            value is another dictionary with a 'data' key which contains the
            string version of the row data for the matched team, and a 'rank'
            key which is the rank of the team.

        Returns
        -------
        dictionary
            An updated version of the team_data_dict with the passed table row
            information included.
        """
        # Teams are listed in terms of rank with the first team being #1
        rank = 1
        for team_data in teams_list:
            abbr = utils._parse_field(PARSING_SCHEME,
                                      team_data,
                                      'abbreviation')
            try:
                team_data_dict[abbr]['data'] += team_data
            except KeyError:
                team_data_dict[abbr] = {'data': team_data, 'rank': rank}
            rank += 1
        return team_data_dict

    def _retrieve_all_teams(self, year):
        """
        Find and create Team instances for all teams in the given season.

        For a given season, parses the specified NBA stats table and finds all
        requested stats. Each team then has a Team instance created which
        includes all requested stats and a few identifiers, such as the team's
        name and abbreviation. All of the individual Team instances are added
        to a list.

        Note that this method is called directly once Teams is invoked and does
        not need to be called manually.

        Parameters
        ----------
        year : string
            The requested year to pull stats from.
        """
        team_data_dict = {}

        if not year:
            year = utils._find_year_for_season('nba')
        doc = pq(SEASON_PAGE_URL % year)
        teams_list = utils._get_stats_table(doc, 'div#all_team-stats-base')
        opp_teams_list = utils._get_stats_table(doc,
                                                'div#all_opponent-stats-base')
        for stats_list in [teams_list, opp_teams_list]:
            team_data_dict = self._add_stats_data(stats_list, team_data_dict)

        for team_data in team_data_dict.values():
            team = Team(team_data['data'], team_data['rank'], year)
            self._teams.append(team)

    @property
    def dataframes(self):
        """
        Returns a pandas DataFrame where each row is a representation of the
        Team class. Rows are indexed by the team abbreviation.
        """
        frames = []
        for team in self.__iter__():
            frames.append(team.dataframe)
        return pd.concat(frames)



import pandas as pd
import re
from functools import wraps
from lxml.etree import ParserError, XMLSyntaxError
from pyquery import PyQuery as pq
from .. import utils
from .constants import (NATIONALITY,
                        PLAYER_ELEMENT_INDEX,
                        PLAYER_SCHEME,
                        PLAYER_URL,
                        ROSTER_URL)
from .player import AbstractPlayer
from six.moves.urllib.error import HTTPError


def _cleanup(prop):
    try:
        prop = prop.replace('%', '')
        prop = prop.replace('$', '')
        prop = prop.replace(',', '')
        return prop.replace('+', '')
    # Occurs when a value is of Nonetype. When that happens, return a blank
    # string as whatever came in had an incomplete value.
    except AttributeError:
        return ''


def _int_property_decorator(func):
    @property
    @wraps(func)
    def wrapper(*args):
        index = args[0]._index
        prop = func(*args)
        element_ind = 0
        if func.__name__ in PLAYER_ELEMENT_INDEX.keys():
            element_ind = PLAYER_ELEMENT_INDEX[func.__name__]
        try:
            value = _cleanup(prop[index][element_ind])
            return int(value)
        except (ValueError, TypeError, IndexError):
            # If there is no value, default to None
            return None
    return wrapper


def _float_property_decorator(func):
    @property
    @wraps(func)
    def wrapper(*args):
        index = args[0]._index
        prop = func(*args)
        element_ind = 0
        try:
            value = _cleanup(prop[index][element_ind])
            return float(value)
        except (ValueError, TypeError, IndexError):
            # If there is no value, default to None
            return None
    return wrapper


def _most_recent_decorator(func):
    @property
    @wraps(func)
    def wrapper(*args):
        season = args[0]._most_recent_season
        seasons = args[0]._season
        index = seasons.index(season)
        prop = func(*args)
        element_ind = 0
        try:
            return prop[index][element_ind]
        except (TypeError, IndexError):
            # If there is no value, default to None
            return None
    return wrapper


class Player(AbstractPlayer):
    """
    Get player information and stats for all seasons.

    Given a player ID, such as 'altuvjo01' for Jose Altuve, capture all
    relevant stats and information like name, nationality, height/weight,
    career home runs, last season's batting average, salary, contract amount,
    and much more.

    By default, the class instance will return the player's career stats, but
    single-season stats can be found by calling the instance with the requested
    season as denoted on baseball-reference.com.

    Parameters
    ----------
    player_id : string
        A player's ID according to basketball-reference.com, such as
        'altuvjo01' for Jose Altuve. The player ID can be found by navigating
        to the player's stats page and getting the string between the final
        slash and the '.html' in the URL. In general, the ID is in the format
        'LLLLLFFNN' where 'LLLLL' are the first 5 letters in the player's last
        name, 'FF', are the first 2 letters in the player's first name, and
        'NN' is a number starting at '01' for the first time that player ID has
        been used and increments by 1 for every successive player.
    """
    def __init__(self, player_id):
        self._most_recent_season = ''
        self._index = None
        self._player_id = player_id
        self._season = None
        self._name = None
        self._team_abbreviation = None
        self._position = None
        self._height = None
        self._weight = None
        self._birth_date = None
        self._nationality = None
        self._contract = None
        self._games = None
        self._games_started = None
        self._plate_appearances = None
        self._at_bats = None
        self._runs = None
        self._hits = None
        self._doubles = None
        self._triples = None
        self._home_runs = None
        self._runs_batted_in = None
        self._stolen_bases = None
        self._times_caught_stealing = None
        self._bases_on_balls = None
        self._times_struck_out = None
        self._batting_average = None
        self._on_base_percentage = None
        self._slugging_percentage = None
        self._on_base_plus_slugging_percentage = None
        self._on_base_plus_slugging_percentage_plus = None
        self._total_bases = None
        self._grounded_into_double_plays = None
        self._times_hit_by_pitch = None
        self._sacrifice_hits = None
        self._sacrifice_flies = None
        self._intentional_bases_on_balls = None
        self._complete_games = None
        self._innings_played = None
        self._defensive_chances = None
        self._putouts = None
        self._assists = None
        self._errors = None
        self._double_plays_turned = None
        self._fielding_percentage = None
        self._total_fielding_runs_above_average = None
        self._defensive_runs_saved_above_average = None
        self._total_fielding_runs_above_average_per_innings = None
        self._defensive_runs_saved_above_average_per_innings = None
        self._range_factor_per_nine_innings = None
        self._range_factor_per_game = None
        self._league_fielding_percentage = None
        self._league_range_factor_per_nine_innings = None
        self._league_range_factor_per_game = None
        self._games_in_batting_order = None
        self._games_in_defensive_lineup = None
        self._games_pitcher = None
        self._games_catcher = None
        self._games_first_baseman = None
        self._games_second_baseman = None
        self._games_third_baseman = None
        self._games_shortstop = None
        self._games_left_fielder = None
        self._games_center_fielder = None
        self._games_right_fielder = None
        self._games_outfielder = None
        self._games_designated_hitter = None
        self._games_pinch_hitter = None
        self._games_pinch_runner = None
        # Stats specific to pitchers
        self._wins = None
        self._losses = None
        self._win_percentage = None
        self._era = None
        self._games_finished = None
        self._shutouts = None
        self._saves = None
        self._hits_allowed = None
        self._runs_allowed = None
        self._earned_runs_allowed = None
        self._home_runs_allowed = None
        self._bases_on_balls_given = None
        self._intentional_bases_on_balls_given = None
        self._strikeouts = None
        self._times_hit_player = None
        self._balks = None
        self._wild_pitches = None
        self._batters_faced = None
        self._era_plus = None
        self._fielding_independent_pitching = None
        self._whip = None
        self._hits_against_per_nine_innings = None
        self._home_runs_against_per_nine_innings = None
        self._bases_on_balls_given_per_nine_innings = None
        self._batters_struckout_per_nine_innings = None
        self._strikeouts_thrown_per_walk = None

        player_data = self._pull_player_data()
        self._find_initial_index()
        AbstractPlayer.__init__(self, player_id, self._name, player_data)

    def _build_url(self):
        """
        Create the player's URL to pull stats from.

        The player's URL requires the first letter of the player's last name
        followed by the player ID.

        Returns
        -------
        string
            The string URL for the player's stats page.
        """
        # The first letter of the player's last name is used to sort the player
        # list and is a part of the URL.
        first_character = self._player_id[0]
        return PLAYER_URL % (first_character, self._player_id)

    def _retrieve_html_page(self):
        """
        Download the requested player's stats page.

        Download the requested page and strip all of the comment tags before
        returning a pyquery object which will be used to parse the data.

        Returns
        -------
        PyQuery object
            The requested page is returned as a queriable PyQuery object with
            the comment tags removed.
        """
        url = self._build_url()
        try:
            url_data = pq(url)
        except HTTPError:
            return None
        return pq(utils._remove_html_comment_tags(url_data))

    def _parse_season(self, row):
        """
        Parse the season string from the table.

        The season is generally located in the first column of the stats tables
        and should be parsed to denote which season metrics are being pulled
        from.

        Parameters
        ----------
        row : PyQuery object
            A PyQuery object of a single row in a stats table.

        Returns
        -------
        string
            A string representation of the season in the format 'YYYY', such as
            '2017'.
        """
        return utils._parse_field(PLAYER_SCHEME, row, 'season')

    def _combine_season_stats(self, table_rows, career_stats, all_stats_dict):
        """
        Combine all stats for each season.

        Since all of the stats are spread across multiple tables, they should
        be combined into a single field which can be used to easily query stats
        at once.

        Parameters
        ----------
        table_rows : generator
            A generator where each element is a row in a stats table.
        career_stats : generator
            A generator where each element is a row in the footer of a stats
            table. Career stats are kept in the footer, hence the usage.
        all_stats_dict : dictionary
            A dictionary of all stats separated by season where each key is the
            season ``string``, such as '2017', and the value is a
            ``dictionary`` with a ``string`` of 'data' and ``string``
            containing all of the data.

        Returns
        -------
        dictionary
            Returns an updated version of the passed all_stats_dict which
            includes more metrics from the provided table.
        """
        most_recent_season = self._most_recent_season
        if not table_rows:
            table_rows = []
        for row in table_rows:
            # For now, remove minor-league stats
            if 'class="minors_table hidden"' in str(row) or \
               'class="spacer partial_table"' in str(row) or \
               'class="partial_table"' in str(row):
                continue
            season = self._parse_season(row)
            try:
                all_stats_dict[season]['data'] += str(row)
            except KeyError:
                all_stats_dict[season] = {'data': str(row)}
            most_recent_season = season
        self._most_recent_season = most_recent_season
        if not career_stats:
            return all_stats_dict
        try:
            all_stats_dict['Career']['data'] += str(next(career_stats))
        except KeyError:
            all_stats_dict['Career'] = {'data': str(next(career_stats))}
        return all_stats_dict

    def _combine_all_stats(self, player_info):
        """
        Pull stats from all tables into single data structure.

        Pull the stats from all of the requested tables into a dictionary that
        is separated by season to allow easy queries of the player's stats for
        each season.

        Parameters
        ----------
        player_info : PyQuery object
            A PyQuery object containing all of the stats information for the
            requested player.

        Returns
        -------
        dictionary
            Returns a dictionary where all stats from each table are combined
            by season to allow easy queries by year.
        """
        all_stats_dict = {}

        for table_id in ['batting_standard', 'standard_fielding',
                         'appearances', 'pitching_standard']:
            table_items = utils._get_stats_table(player_info,
                                                 'table#%s' % table_id)
            career_items = utils._get_stats_table(player_info,
                                                  'table#%s' % table_id,
                                                  footer=True)
            all_stats_dict = self._combine_season_stats(table_items,
                                                        career_items,
                                                        all_stats_dict)
        return all_stats_dict

    def _parse_nationality(self, player_info):
        """
        Parse the player's nationality.

        The player's nationality is denoted by a flag in the information
        section with a country code for each nation. The country code needs to
        pulled and then matched to find the player's home country. Once found,
        the '_nationality' attribute is set for the player.

        Parameters
        ----------
        player_info : PyQuery object
            A PyQuery object containing the HTML from the player's stats page.
        """
        for span in player_info('span').items():
            if 'class="f-i' in str(span):
                nationality = span.text()
                nationality = NATIONALITY[nationality]
                setattr(self, '_nationality', nationality)
                break

    def _parse_player_information(self, player_info):
        """
        Parse general player information.

        Parse general player information such as height, weight, and name. The
        attribute for the requested field will be set with the value prior to
        returning.

        Parameters
        ----------
        player_info : PyQuery object
            A PyQuery object containing the HTML from the player's stats page.
        """
        for field in ['_height', '_weight', '_name']:
            short_field = str(field)[1:]
            value = utils._parse_field(PLAYER_SCHEME, player_info, short_field)
            setattr(self, field, value)

    def _parse_birth_date(self, player_info):
        """
        Parse the player's birth date.

        Pull the player's birth date from the player information and set the
        '_birth_date' attribute with the value prior to returning.

        Parameters
        ----------
        player_info : PyQuery object
            A PyQuery object containing the HTML from the player's stats page.
        """
        date = player_info('span[itemprop="birthDate"]').attr('data-birth')
        setattr(self, '_birth_date', date)

    def _parse_team_name(self, team):
        """
        Parse the team name in the contract table.

        The team names in the contract table contain special encoded characters
        that are not supported by Python 2.7. These characters should be
        filtered out to get the proper team name.

        Parameters
        ----------
        team : string
            A string representing the team_name tag in a row in the player's
            contract table.

        Returns
        -------
        string
            A string of the team's name, such as 'Houston Astros'.
        """
        team = team.replace('&#160;', ' ')
        team = team.replace('\xa0', ' ')
        team_html = pq(team)
        return team_html.text()

    def _parse_contract(self, player_info):
        """
        Parse the player's contract.

        Depending on the player's contract status, a contract table is located
        at the bottom of the stats page and includes player wages by season. If
        found, create a dictionary housing the wages by season.

        Parameters
        ----------
        player_info : PyQuery object
            A PyQuery object containing the HTML from the player's stats page.
        """
        contract = {}

        salary_table = player_info('table#br-salaries')
        for row in salary_table('tbody tr').items():
            if 'class="spacer partial_table"' in str(row):
                continue
            year = row('th[data-stat="year_ID"]').text()
            if year.strip() == '':
                continue
            age = row('td[data-stat="age"]').text()
            team = self._parse_team_name(str(row('td[data-stat="team_name"]')))
            salary = row('td[data-stat="Salary"]').text()
            contract[year] = {
                'age': age,
                'team': team,
                'salary': salary
            }
        setattr(self, '_contract', contract)

    def _parse_value(self, html_data, field):
        """
        Parse the HTML table to find the requested field's value.

        All of the values are passed in an HTML table row instead of as
        individual items. The values need to be parsed by matching the
        requested attribute with a parsing scheme that sports-reference uses
        to differentiate stats. This function returns a single value for the
        given attribute.

        Parameters
        ----------
        html_data : string
            A string containing all of the rows of stats for a given team. If
            multiple tables are being referenced, this will be comprised of
            multiple rows in a single string.
        field : string
            The name of the attribute to match. Field must be a key in the
            PLAYER_SCHEME dictionary.

        Returns
        -------
        list
            A list of all values that match the requested field. If no value
            could be found, returns None.
        """
        scheme = PLAYER_SCHEME[field]
        items = [i.text() for i in html_data(scheme).items()]
        # Stats can be added and removed on a yearly basis. If no stats are
        # found, return None and have that be the value.
        if len(items) == 0:
            return None
        return items

    def _pull_player_data(self):
        """
        Pull and aggregate all player information.

        Pull the player's HTML stats page and parse unique properties, such as
        the player's height, weight, and position. Next, combine all stats for
        all seasons plus the player's career stats into a single object which
        can easily be iterated upon.

        Returns
        -------
        dictionary
            Returns a dictionary of the player's combined stats where each key
            is a string of the season and the value is the season's associated
            stats.
        """
        player_info = self._retrieve_html_page()
        self._parse_player_information(player_info)
        self._parse_nationality(player_info)
        self._parse_birth_date(player_info)
        self._parse_contract(player_info)
        all_stats = self._combine_all_stats(player_info)
        setattr(self, '_season', list(all_stats.keys()))
        return all_stats

    def _find_initial_index(self):
        """
        Find the index of career stats.

        When the Player class is instantiated, the default stats to pull are
        the player's career stats. Upon being called, the index of the 'Career'
        element should be the index value.
        """
        index = 0
        for season in self._season:
            # The career stats default to Nonetype
            if season is None or season == 'Career':
                self._index = index
                self._season[index] = 'Career'
                break
            index += 1

    def __call__(self, requested_season=''):
        """
        Specify a different season to pull stats from.

        A different season can be requested by passing the season string, such
        as '2017' to the class instance.

        Parameters
        ----------
        requested_season : string (optional)
            A string of the requested season to query, such as '2017'. If left
            blank or 'Career' is passed, the career stats will be used for
            stats queries.

        Returns
        -------
        Player class instance
            Returns the class instance with the updated stats being referenced.
        """
        if requested_season.lower() == 'career' or \
           requested_season == '':
            requested_season = 'Career'
        index = 0
        for season in self._season:
            if season == requested_season:
                self._index = index
                break
            index += 1
        return self

    def _dataframe_fields(self):
        """
        Creates a dictionary of all fields to include with DataFrame.

        With the result of the calls to class properties changing based on the
        class index value, the dictionary should be regenerated every time the
        index is changed when the dataframe property is requested.

        Returns
        -------
        dictionary
            Returns a dictionary where the keys are the shortened ``string``
            attribute names and the values are the actual value for each
            attribute for the specified index.
        """
        fields_to_include = {
            'assists': self.assists,
            'at_bats': self.at_bats,
            'bases_on_balls': self.bases_on_balls,
            'batting_average': self.batting_average,
            'birth_date': self.birth_date,
            'complete_games': self.complete_games,
            'defensive_chances': self.defensive_chances,
            'defensive_runs_saved_above_average':
            self.defensive_runs_saved_above_average,
            'defensive_runs_saved_above_average_per_innings':
            self.defensive_runs_saved_above_average_per_innings,
            'double_plays_turned': self.double_plays_turned,
            'doubles': self.doubles,
            'errors': self.errors,
            'fielding_percentage': self.fielding_percentage,
            'games': self.games,
            'games_catcher': self.games_catcher,
            'games_center_fielder': self.games_center_fielder,
            'games_designated_hitter': self.games_designated_hitter,
            'games_first_baseman': self.games_first_baseman,
            'games_in_batting_order': self.games_in_batting_order,
            'games_in_defensive_lineup': self.games_in_defensive_lineup,
            'games_left_fielder': self.games_left_fielder,
            'games_outfielder': self.games_outfielder,
            'games_pinch_hitter': self.games_pinch_hitter,
            'games_pinch_runner': self.games_pinch_runner,
            'games_pitcher': self.games_pitcher,
            'games_right_fielder': self.games_right_fielder,
            'games_second_baseman': self.games_second_baseman,
            'games_shortstop': self.games_shortstop,
            'games_started': self.games_started,
            'games_third_baseman': self.games_third_baseman,
            'grounded_into_double_plays': self.grounded_into_double_plays,
            'height': self.height,
            'hits': self.hits,
            'home_runs': self.home_runs,
            'innings_played': self.innings_played,
            'intentional_bases_on_balls': self.intentional_bases_on_balls,
            'league_fielding_percentage': self.league_fielding_percentage,
            'league_range_factor_per_game': self.league_range_factor_per_game,
            'league_range_factor_per_nine_innings':
            self.league_range_factor_per_nine_innings,
            'name': self.name,
            'nationality': self.nationality,
            'on_base_percentage': self.on_base_percentage,
            'on_base_plus_slugging_percentage':
            self.on_base_plus_slugging_percentage,
            'on_base_plus_slugging_percentage_plus':
            self.on_base_plus_slugging_percentage_plus,
            'plate_appearances': self.plate_appearances,
            'player_id': self.player_id,
            'position': self.position,
            'putouts': self.putouts,
            'range_factor_per_game': self.range_factor_per_game,
            'range_factor_per_nine_innings':
            self.range_factor_per_nine_innings,
            'runs': self.runs,
            'runs_batted_in': self.runs_batted_in,
            'sacrifice_flies': self.sacrifice_flies,
            'sacrifice_hits': self.sacrifice_hits,
            'season': self.season,
            'slugging_percentage': self.slugging_percentage,
            'stolen_bases': self.stolen_bases,
            'team_abbreviation': self.team_abbreviation,
            'times_caught_stealing': self.times_caught_stealing,
            'times_hit_by_pitch': self.times_hit_by_pitch,
            'times_struck_out': self.times_struck_out,
            'total_bases': self.total_bases,
            'total_fielding_runs_above_average':
            self.total_fielding_runs_above_average,
            'total_fielding_runs_above_average_per_innings':
            self.total_fielding_runs_above_average_per_innings,
            'triples': self.triples,
            'weight': self.weight,
            # Properties specific to pitchers
            'balks': self.balks,
            'bases_on_balls_given': self.bases_on_balls_given,
            'bases_on_balls_given_per_nine_innings':
            self.bases_on_balls_given_per_nine_innings,
            'batters_faced': self.batters_faced,
            'batters_struckout_per_nine_innings':
            self.batters_struckout_per_nine_innings,
            'earned_runs_allowed': self.earned_runs_allowed,
            'era': self.era,
            'era_plus': self.era_plus,
            'fielding_independent_pitching':
            self.fielding_independent_pitching,
            'games_finished': self.games_finished,
            'hits_against_per_nine_innings':
            self.hits_against_per_nine_innings,
            'hits_allowed': self.hits_allowed,
            'home_runs_against_per_nine_innings':
            self.home_runs_against_per_nine_innings,
            'home_runs_allowed': self.home_runs_allowed,
            'intentional_bases_on_balls_given':
            self.intentional_bases_on_balls_given,
            'losses': self.losses,
            'runs_allowed': self.runs_allowed,
            'saves': self.saves,
            'shutouts': self.shutouts,
            'strikeouts': self.strikeouts,
            'strikeouts_thrown_per_walk': self.strikeouts_thrown_per_walk,
            'times_hit_player': self.times_hit_player,
            'whip': self.whip,
            'wild_pitches': self.wild_pitches,
            'win_percentage': self.win_percentage,
            'wins': self.wins
        }
        return fields_to_include

    @property
    def dataframe(self):
        """
        Returns a ``pandas DataFrame`` containing all other relevant class
        properties and values where each index is a different season plus the
        career stats.
        """
        temp_index = self._index
        rows = []
        indices = []
        for season in self._season:
            self._index = self._season.index(season)
            rows.append(self._dataframe_fields())
            indices.append(season)
        self._index = temp_index
        return pd.DataFrame(rows, index=[indices])

    @property
    def season(self):
        """
        Returns a ``string`` of the season in the format 'YYYY', such as
        '2017'. If no season was requsted, the career stats will be
        returned for the player and the season will default to 'Career'.
        """
        return self._season[self._index]

    @property
    def name(self):
        """
        Returns a ``string`` of the player's name, such as 'Jose Altuve'.
        """
        return self._name

    @_most_recent_decorator
    def team_abbreviation(self):
        """
        Returns a ``string`` of the team's abbreviation, such as 'HOU' for the
        Houston Astros.
        """
        return self._team_abbreviation

    @_most_recent_decorator
    def position(self):
        """
        Returns a ``string`` constant of the player's primary position.
        """
        return self._position

    @property
    def height(self):
        """
        Returns a ``string`` of the players height in the format "feet-inches".
        """
        return self._height

    @property
    def weight(self):
        """
        Returns an ``int`` of the player's weight in pounds.
        """
        return int(self._weight.replace('lb', ''))

    @property
    def birth_date(self):
        """
        Returns a ``datetime`` object of the day and year the player was born.
        """
        return self._birth_date

    @property
    def nationality(self):
        """
        Returns a ``string`` constant denoting which country the player
        originiates from.
        """
        return self._nationality

    @property
    def contract(self):
        """
        Returns a ``dictionary`` of the player's contract where each key is a
        ``string`` of the year, such as '2017' and each value is a
        ``dictionary`` with the ``string`` key-value pairs of the player's age,
        team name, and salary.
        """
        return self._contract

    @_int_property_decorator
    def games(self):
        """
        Returns an ``int`` of the number of games the player participated in.
        """
        return self._games

    @_int_property_decorator
    def games_started(self):
        """
        Returns an ``int`` of the number of games the player started.
        """
        return self._games_started

    @_int_property_decorator
    def plate_appearances(self):
        """
        Returns an ``int`` of the number of plate appearances the player had.
        """
        return self._plate_appearances

    @_int_property_decorator
    def at_bats(self):
        """
        Returns an ``int`` of the number of at bats the player had.
        """
        return self._at_bats

    @_int_property_decorator
    def runs(self):
        """
        Returns an ``int`` of the number of runs the player scored.
        """
        return self._runs

    @_int_property_decorator
    def hits(self):
        """
        Returns an ``int`` of the number of hits the player had.
        """
        return self._hits

    @_int_property_decorator
    def doubles(self):
        """
        Returns an ``int`` of the number of doubles the player hit.
        """
        return self._doubles

    @_int_property_decorator
    def triples(self):
        """
        Returns an ``int`` of the number of triples the player hit.
        """
        return self._triples

    @_int_property_decorator
    def home_runs(self):
        """
        Returns an ``int`` of the number of home runs the player hit.
        """
        return self._home_runs

    @_int_property_decorator
    def runs_batted_in(self):
        """
        Returns an ``int`` of the number of runs batted in the player
        registered.
        """
        return self._runs_batted_in

    @_int_property_decorator
    def stolen_bases(self):
        """
        Returns an ``int`` of the number of bases the player has stolen.
        """
        return self._stolen_bases

    @_int_property_decorator
    def times_caught_stealing(self):
        """
        Returns an ``int`` of the number of times the player was caught
        stealing.
        """
        return self._times_caught_stealing

    @_int_property_decorator
    def bases_on_balls(self):
        """
        Returns an ``int`` of the number of bases the player registered as a
        result of balls.
        """
        return self._bases_on_balls

    @_int_property_decorator
    def times_struck_out(self):
        """
        Returns an ``int`` of the number of times the player was struck out.
        """
        return self._times_struck_out

    @_float_property_decorator
    def batting_average(self):
        """
        Returns a ``float`` of the batting average for the player.
        """
        return self._batting_average

    @_float_property_decorator
    def on_base_percentage(self):
        """
        Returns a ``float`` of the percentage of at bats that result in the
        batter getting on base.
        """
        return self._on_base_percentage

    @_float_property_decorator
    def slugging_percentage(self):
        """
        Returns a ``float`` of the slugging percentage for the player based
        on the number of bases gained per at-bat with bigger plays getting more
        weight.
        """
        return self._slugging_percentage

    @_float_property_decorator
    def on_base_plus_slugging_percentage(self):
        """
        Returns a ``float`` of the on base percentage plus the slugging
        percentage. Percentage ranges from 0-1.
        """
        return self._on_base_plus_slugging_percentage

    @_int_property_decorator
    def on_base_plus_slugging_percentage_plus(self):
        """
        Returns an ``int`` of the on base percentage plus the slugging
        percentage, adjusted to the player's ballpark.
        """
        return self._on_base_plus_slugging_percentage_plus

    @_int_property_decorator
    def total_bases(self):
        """
        Returns an ``int`` of the number of bases the player has gained.
        """
        return self._total_bases

    @_int_property_decorator
    def grounded_into_double_plays(self):
        """
        Returns an ``int`` of the number of double plays the player grounded
        into.
        """
        return self._grounded_into_double_plays

    @_int_property_decorator
    def times_hit_by_pitch(self):
        """
        Returns an ``int`` of the number of times the player has been hit by a
        pitch.
        """
        return self._times_hit_by_pitch

    @_int_property_decorator
    def sacrifice_hits(self):
        """
        Returns an ``int`` of the number of sacrifice hits or sacrafice bunts
        the player made.
        """
        return self._sacrifice_hits

    @_int_property_decorator
    def sacrifice_flies(self):
        """
        Returns an ``int`` of the number of sacrifice flies the player hit.
        """
        return self._sacrifice_flies

    @_int_property_decorator
    def intentional_bases_on_balls(self):
        """
        Returns an ``int`` of the number of times the player has been
        intentionally walked by the opposition.
        """
        return self._intentional_bases_on_balls

    @_int_property_decorator
    def complete_games(self):
        """
        Returns an ``int`` of the number of complete games the player has
        participated in.
        """
        return self._complete_games

    @_float_property_decorator
    def innings_played(self):
        """
        Returns a ``float`` of the total number of innings the player has
        played in.
        """
        return self._innings_played

    @_int_property_decorator
    def defensive_chances(self):
        """
        Returns an ``int`` of the number of defensive chances (equal to the
        number of putouts + assists + errors) the player had.
        """
        return self._defensive_chances

    @_int_property_decorator
    def putouts(self):
        """
        Returns an ``int`` of the number of putouts the player had.
        """
        return self._putouts

    @_int_property_decorator
    def assists(self):
        """
        Returns an ``int`` of the number of assists the player had.
        """
        return self._assists

    @_int_property_decorator
    def errors(self):
        """
        Returns an ``int`` of the number of errors the player made.
        """
        return self._errors

    @_int_property_decorator
    def double_plays_turned(self):
        """
        Returns an ``int`` of the number of double plays the player was
        involved in.
        """
        return self._double_plays_turned

    @_float_property_decorator
    def fielding_percentage(self):
        """
        Returns a ``float`` of the players fielding percentage, equivalent to
        (putouts + assists) / (putouts + assists + errors). Percentage ranges
        from 0-1.
        """
        return self._fielding_percentage

    @_int_property_decorator
    def total_fielding_runs_above_average(self):
        """
        Returns an ``int`` of the number of runs the player was worth compared
        to an average player.
        """
        return self._total_fielding_runs_above_average

    @_int_property_decorator
    def defensive_runs_saved_above_average(self):
        """
        Returns an ``int`` of the number of defensive runs the player saved
        compared to an average player.
        """
        return self._defensive_runs_saved_above_average

    @_int_property_decorator
    def total_fielding_runs_above_average_per_innings(self):
        """
        Returns an ``int`` of the number of runs the player was worth per 1,200
        innings compared to an average player.
        """
        return self._total_fielding_runs_above_average_per_innings

    @_int_property_decorator
    def defensive_runs_saved_above_average_per_innings(self):
        """
        Returns an ``int`` of the number of defensive runs the player was worth
        per 1,200 innings compared to an average player.
        """
        return self._defensive_runs_saved_above_average_per_innings

    @_float_property_decorator
    def range_factor_per_nine_innings(self):
        """
        Returns a ``float`` of the players range factor per nine innings, equal
        to 9 * (putouts + assists) / innings_played.
        """
        return self._range_factor_per_nine_innings

    @_float_property_decorator
    def range_factor_per_game(self):
        """
        Returns a ``float`` of the players range factor per game, equal to 9 *
        (putouts + assists) / games_played.
        """
        return self._range_factor_per_game

    @_float_property_decorator
    def league_fielding_percentage(self):
        """
        Returns a ``float`` of the average fielding percentage for the league
        at the player's position. Percentage ranges from 0-1.
        """
        return self._league_fielding_percentage

    @_float_property_decorator
    def league_range_factor_per_nine_innings(self):
        """
        Returns a ``float`` of the average range factor for the league per nine
        innings, equal to 9 * (putouts + assists) / innings_played.
        """
        return self._league_range_factor_per_nine_innings

    @_float_property_decorator
    def league_range_factor_per_game(self):
        """
        Returns a ``float`` of the average range factor for the league per
        game, equal to (putouts + assists) / games_played.
        """
        return self._league_range_factor_per_game

    @_int_property_decorator
    def games_in_batting_order(self):
        """
        Returns an ``int`` of the number of games the player was in the batting
        lineup.
        """
        return self._games_in_batting_order

    @_int_property_decorator
    def games_in_defensive_lineup(self):
        """
        Returns an ``int`` of the number of games the player was in the
        defensive lineup.
        """
        return self._games_in_defensive_lineup

    @_int_property_decorator
    def games_pitcher(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a pitcher.
        """
        return self._games_pitcher

    @_int_property_decorator
    def games_catcher(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a catcher.
        """
        return self._games_catcher

    @_int_property_decorator
    def games_first_baseman(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a first baseman.
        """
        return self._games_first_baseman

    @_int_property_decorator
    def games_second_baseman(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a second baseman.
        """
        return self._games_second_baseman

    @_int_property_decorator
    def games_third_baseman(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a third baseman.
        """
        return self._games_third_baseman

    @_int_property_decorator
    def games_shortstop(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a shortstop.
        """
        return self._games_shortstop

    @_int_property_decorator
    def games_left_fielder(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a left fielder.
        """
        return self._games_left_fielder

    @_int_property_decorator
    def games_center_fielder(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a center fielder.
        """
        return self._games_center_fielder

    @_int_property_decorator
    def games_right_fielder(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a right fielder.
        """
        return self._games_right_fielder

    @_int_property_decorator
    def games_outfielder(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as an outfielder.
        """
        return self._games_outfielder

    @_int_property_decorator
    def games_designated_hitter(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a designated hitter.
        """
        return self._games_designated_hitter

    @_int_property_decorator
    def games_pinch_hitter(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a pinch hitter.
        """
        return self._games_pinch_hitter

    @_int_property_decorator
    def games_pinch_runner(self):
        """
        Returns an ``int`` of the number of games the player was in the lineup
        as a pinch runner.
        """
        return self._games_pinch_runner

    @_int_property_decorator
    def wins(self):
        """
        Returns an ``int`` of the number of games the player has won as a
        pitcher.
        """
        return self._wins

    @_int_property_decorator
    def losses(self):
        """
        Returns an ``int`` of the number of games the player has lost as a
        pitcher.
        """
        return self._losses

    @_float_property_decorator
    def win_percentage(self):
        """
        Returns a ``float`` of the players winning percentage as a pitcher.
        Percentage ranges from 0-1.
        """
        return self._win_percentage

    @_float_property_decorator
    def era(self):
        """
        Returns a ``float`` of the pitcher's Earned Runs Average.
        """
        return self._era

    @_int_property_decorator
    def games_finished(self):
        """
        Returns an ``int`` of the number of games the player finished as a
        pitcher.
        """
        return self._games_finished

    @_int_property_decorator
    def shutouts(self):
        """
        Returns an ``int`` of the number of times the player did not allow any
        runs and threw a complete game as a pitcher.
        """
        return self._shutouts

    @_int_property_decorator
    def saves(self):
        """
        Returns an ``int`` of the number of saves the player made as a pitcher.
        """
        return self._saves

    @_int_property_decorator
    def hits_allowed(self):
        """
        Returns an ``int`` of the number of hits the player allowed as a
        pitcher.
        """
        return self._hits_allowed

    @_int_property_decorator
    def runs_allowed(self):
        """
        Returns an ``int`` of the number of runs the player allowed as a
        pitcher.
        """
        return self._runs_allowed

    @_int_property_decorator
    def earned_runs_allowed(self):
        """
        Returns an ``int`` of the number of earned runs the player allowed as a
        pitcher.
        """
        return self._earned_runs_allowed

    @_int_property_decorator
    def home_runs_allowed(self):
        """
        Returns an ``int`` of the number of home runs a player has allowed as a
        pitcher.
        """
        return self._home_runs_allowed

    @_int_property_decorator
    def bases_on_balls_given(self):
        """
        Returns an ``int`` of the number of bases on balls the player has given
        as a pitcher.
        """
        return self._bases_on_balls_given

    @_int_property_decorator
    def intentional_bases_on_balls_given(self):
        """
        Returns an ``int`` of the number of bases the player has intentionally
        given as a pitcher.
        """
        return self._intentional_bases_on_balls_given

    @_int_property_decorator
    def strikeouts(self):
        """
        Returns an ``int`` of the number of strikeouts the player threw as a
        pitcher.
        """
        return self._strikeouts

    @_int_property_decorator
    def times_hit_player(self):
        """
        Returns an ``int`` of the number of times the pitcher hit a player with
        a pitch.
        """
        return self._times_hit_player

    @_int_property_decorator
    def balks(self):
        """
        Returns an ``int`` of the number of times the pitcher balked.
        """
        return self._balks

    @_int_property_decorator
    def wild_pitches(self):
        """
        Returns an ``int`` of the number of wild pitches the player has thrown.
        """
        return self._wild_pitches

    @_float_property_decorator
    def era_plus(self):
        """
        Returns a ``float`` of the pitcher's ERA while adjusted for the
        ballpark.
        """
        return self._era_plus

    @_float_property_decorator
    def fielding_independent_pitching(self):
        """
        Returns a ``float`` of the pitcher's effectiveness at preventing home
        runs, bases on balls, and hitting players with pitches, while causing
        strikeouts.
        """
        return self._fielding_independent_pitching

    @_float_property_decorator
    def whip(self):
        """
        Returns a ``float`` of the pitcher's WHIP score, equivalent to (bases
        on balls + hits) / innings played.
        """
        return self._whip

    @_float_property_decorator
    def hits_against_per_nine_innings(self):
        """
        Returns a ``float`` of the number of hits the player has given per nine
        innings played.
        """
        return self._hits_against_per_nine_innings

    @_float_property_decorator
    def home_runs_against_per_nine_innings(self):
        """
        Returns a ``float`` of the number of home runs the pitcher has given
        per nine innings played.
        """
        return self._home_runs_against_per_nine_innings

    @_float_property_decorator
    def bases_on_balls_given_per_nine_innings(self):
        """
        Returns a ``float`` of the number of bases on balls the pitcher has
        given per nine innings played.
        """
        return self._bases_on_balls_given_per_nine_innings

    @_float_property_decorator
    def batters_struckout_per_nine_innings(self):
        """
        Returns a ``float`` of the number of batters the pitcher has struck out
        per nine innings played.
        """
        return self._batters_struckout_per_nine_innings

    @_float_property_decorator
    def strikeouts_thrown_per_walk(self):
        """
        Returns a ``float`` of the number of batters the pitcher has struck out
        per the number of walks given.
        """
        return self._strikeouts_thrown_per_walk


class Roster(object):
    """
    Get stats for all players on a roster.

    Request a team's roster for a given season and create instances of the
    Player class for each player, containing a detailed list of the players
    statistics and information.

    Parameters
    ----------
    team : string
        The team's abbreviation, such as 'HOU' for the Houston Astros.
    year : string (optional)
        The 4-digit year to pull the roster from, such as '2018'. If left
        blank, defaults to the most recent season.
    slim : boolean (optional)
        Set to True to return a limited subset of player information including
        the name and player ID for each player as opposed to all of their
        respective stats which greatly reduces the time to return a response if
        just the names and IDs are desired. Defaults to False.
    """
    def __init__(self, team, year=None, slim=False):
        self._team = team
        self._slim = slim
        if slim:
            self._players = {}
        else:
            self._players = []

        self._find_players(year)

    def _pull_team_page(self, url):
        """
        Download the team page.

        Download the requested team's season page and create a PyQuery object.

        Parameters
        ----------
        url : string
            A string of the built URL for the requested team and season.

        Returns
        -------
        PyQuery object
            Returns a PyQuery object of the team's HTML page.
        """
        try:
            return pq(url)
        except HTTPError:
            return None

    def _create_url(self, year):
        """
        Build the team URL.

        Build a URL given a team's abbreviation and the 4-digit year.

        Parameters
        ----------
        year : string
            The 4-digit string representing the year to pull the team's roster
            from.

        Returns
        -------
        string
            Returns a string of the team's season page for the requested team
            and year.
        """
        return ROSTER_URL % (self._team.upper(), year)

    def _get_id(self, player):
        """
        Parse the player ID.

        Given a PyQuery object representing a single player on the team roster,
        parse the player ID and return it as a string.

        Parameters
        ----------
        player : PyQuery object
            A PyQuery object representing the player information from the
            roster table.

        Returns
        -------
        string
            Returns a string of the player ID.
        """
        name_tag = player('td[data-stat="player"] a')
        name = re.sub(r'.*/players/./', '', str(name_tag))
        return re.sub(r'\.shtml.*', '', name)

    def _get_name(self, player):
        """
        Parse the player's name.

        Given a PyQuery object representing a single player on the team roster,
        parse the player ID and return it as a string.

        Parameters
        ----------
        player : PyQuery object
            A PyQuery object representing the player information from the
            roster table.

        Returns
        -------
        string
            Returns a string of the player's name.
        """
        name_tag = player('td[data-stat="player"] a')
        return name_tag.text()

    def _find_players(self, year):
        """
        Find all player IDs for the requested team.

        For the requested team and year (if applicable), pull the roster table
        and parse the player ID for all players on the roster and create an
        instance of the Player class for the player. All player instances are
        added to the 'players' property to get all stats for all players on a
        team.

        Parameters
        ----------
        year : string
            The 4-digit string representing the year to pull the team's roster
            from.
        """
        if not year:
            year = utils._find_year_for_season('mlb')
        url = self._create_url(year)
        page = self._pull_team_page(url)
        if not page:
            output = ("Can't pull requested team page. Ensure the following "
                      "URL exists: %s" % url)
            raise ValueError(output)
        players = page('table#team_batting tbody tr').items()
        players_parsed = []
        for player in players:
            if 'class="thead"' in str(player):
                continue
            player_id = self._get_id(player)
            if self._slim:
                name = self._get_name(player)
                self._players[player_id] = name
            else:
                player_instance = Player(player_id)
                self._players.append(player_instance)
            players_parsed.append(player_id)
        for player in page('table#team_pitching tbody tr').items():
            if 'class="thead"' in str(player):
                continue
            player_id = self._get_id(player)
            # Skip players that showup in both batting and pitching tables, as
            # is often the case with National League pitchers.
            if player_id in players_parsed:
                continue
            if self._slim:
                name = self._get_name(player)
                self._players[player_id] = name
            else:
                player_instance = Player(player_id)
                self._players.append(player_instance)

    @property
    def players(self):
        """
        Returns a ``list`` of player instances for each player on the requested
        team's roster if the ``slim`` property is False when calling the Roster
        class. If the ``slim`` property is True, returns a ``dictionary`` where
        each key is a string of the player's ID and each value is the player's
        first and last name as listed on the roster page.
        """
        return self._players



import pandas as pd
import re
from functools import wraps
from lxml.etree import ParserError, XMLSyntaxError
from pyquery import PyQuery as pq
from .. import utils
from .constants import (BOXSCORE_SCHEME,
                        NATIONALITY,
                        PLAYER_ELEMENT_INDEX,
                        PLAYER_SCHEME,
                        PLAYER_URL,
                        ROSTER_URL)
from six.moves.urllib.error import HTTPError


def _cleanup(prop):
    try:
        prop = prop.replace('%', '')
        prop = prop.replace('$', '')
        prop = prop.replace(',', '')
        return prop.replace('+', '')
    # Occurs when a value is of Nonetype. When that happens, return a blank
    # string as whatever came in had an incomplete value.
    except AttributeError:
        return ''


def _int_property_decorator(func):
    @property
    @wraps(func)
    def wrapper(*args):
        index = args[0]._index
        prop = func(*args)
        element_ind = 0
        if func.__name__ in PLAYER_ELEMENT_INDEX.keys():
            element_ind = PLAYER_ELEMENT_INDEX[func.__name__]
        try:
            value = _cleanup(prop[index][element_ind])
            return int(value)
        except (ValueError, TypeError, IndexError):
            # If there is no value, default to None
            return None
    return wrapper


def _float_property_decorator(func):
    @property
    @wraps(func)
    def wrapper(*args):
        index = args[0]._index
        prop = func(*args)
        element_ind = 0
        if func.__name__ in PLAYER_ELEMENT_INDEX.keys():
            element_ind = PLAYER_ELEMENT_INDEX[func.__name__]
        try:
            value = _cleanup(prop[index][element_ind])
            return float(value)
        except (ValueError, TypeError, IndexError):
            # If there is no value, default to None
            return None
    return wrapper


class AbstractPlayer(object):
    """
    Get player information and stats for all seasons.

    Given a player ID, such as 'altuvjo01' for Jose Altuve, capture all
    relevant stats and information like name, nationality, height/weight,
    career home runs, last season's batting average, salary, contract amount,
    and much more.

    By default, the class instance will return the player's career stats, but
    single-season stats can be found by calling the instance with the requested
    season as denoted on baseball-reference.com.

    Parameters
    ----------
    player_id : string
        A player's ID according to basketball-reference.com, such as
        'altuvjo01' for Jose Altuve. The player ID can be found by navigating
        to the player's stats page and getting the string between the final
        slash and the '.html' in the URL. In general, the ID is in the format
        'LLLLLFFNN' where 'LLLLL' are the first 5 letters in the player's last
        name, 'FF', are the first 2 letters in the player's first name, and
        'NN' is a number starting at '01' for the first time that player ID has
        been used and increments by 1 for every successive player.
    """
    def __init__(self, player_id, player_name, player_data):
        self._player_id = player_id
        self._name = player_name
        self._plate_appearances = None
        self._at_bats = None
        self._runs = None
        self._hits = None
        self._runs_batted_in = None
        self._bases_on_balls = None
        self._times_struck_out = None
        self._batting_average = None
        self._on_base_percentage = None
        self._slugging_percentage = None
        self._on_base_plus_slugging_percentage = None
        self._putouts = None
        self._assists = None
        # Stats specific to pitchers
        self._hits_allowed = None
        self._runs_allowed = None
        self._earned_runs_allowed = None
        self._home_runs_allowed = None
        self._bases_on_balls_given = None
        self._strikeouts = None
        self._batters_faced = None

        self._parse_player_data(player_data)

    def _parse_value(self, stats, field):
        """
        Pull the specified value from the HTML contents.

        Given a field, find the corresponding HTML tag for that field and parse
        its value before returning the value as a string.

        Parameters
        ----------
        stats : PyQuery object
            A PyQuery object containing all stats in HTML format for a
            particular player.
        field : string
            A string of the field to parse from the HTML.

        Returns
        -------
        list
            A list of all values that match the requested field. If no value
            could be found, returns None.
        """
        scheme = PLAYER_SCHEME[field]
        items = [i.text() for i in stats(scheme).items()]
        # Stats can be added and removed on a yearly basis. If no stats are
        # found, return None and have that be the value.
        if len(items) == 0:
            return None
        return items

    def _parse_player_data(self, player_data):
        """
        Parse all player information and set attributes.

        Iterate through each class attribute to parse the data from the HTML
        page and set the attribute value with the result.

        Parameters
        ----------
        player_data : dictionary or string
            If this class is inherited from the ``Player`` class, player_data
            will be a dictionary where each key is a string representing the
            season and each value contains the HTML data as a string. If this
            class is inherited from the ``BoxscorePlayer`` class, player_data
            will be a string representing the player's game statistics in HTML
            format.
        """
        for field in self.__dict__:
            short_field = str(field)[1:]
            if short_field == 'player_id' or \
               short_field == 'index' or \
               short_field == 'most_recent_season' or \
               short_field == 'season' or \
               short_field == 'name' or \
               short_field == 'weight' or \
               short_field == 'height' or \
               short_field == 'nationality' or \
               short_field == 'birth_date' or \
               short_field == 'contract':
                continue
            field_stats = []
            if type(player_data) == dict:
                for year, data in player_data.items():
                    stats = pq(data['data'])
                    value = self._parse_value(stats, short_field)
                    field_stats.append(value)
            else:
                stats = pq(player_data)
                value = self._parse_value(stats, short_field)
                field_stats.append(value)
            setattr(self, field, field_stats)

    @property
    def player_id(self):
        """
        Returns a ``string`` of the player's ID on sports-reference, such as
        'altuvjo01' for Jose Altuve.
        """
        return self._player_id

    @property
    def name(self):
        """
        Returns a ``string`` of the player's name, such as 'Jose Altuve'.
        """
        return self._name

    @_int_property_decorator
    def plate_appearances(self):
        """
        Returns an ``int`` of the number of plate appearances the player had.
        """
        return self._plate_appearances

    @_int_property_decorator
    def at_bats(self):
        """
        Returns an ``int`` of the number of at bats the player had.
        """
        return self._at_bats

    @_int_property_decorator
    def runs(self):
        """
        Returns an ``int`` of the number of runs the player scored.
        """
        return self._runs

    @_int_property_decorator
    def hits(self):
        """
        Returns an ``int`` of the number of hits the player had.
        """
        return self._hits

    @_int_property_decorator
    def runs_batted_in(self):
        """
        Returns an ``int`` of the number of runs batted in the player
        registered.
        """
        return self._runs_batted_in

    @_int_property_decorator
    def bases_on_balls(self):
        """
        Returns an ``int`` of the number of bases the player registered as a
        result of balls.
        """
        return self._bases_on_balls

    @_int_property_decorator
    def times_struck_out(self):
        """
        Returns an ``int`` of the number of times the player was struck out.
        """
        return self._times_struck_out

    @_float_property_decorator
    def batting_average(self):
        """
        Returns a ``float`` of the batting average for the player.
        """
        return self._batting_average

    @_float_property_decorator
    def on_base_percentage(self):
        """
        Returns a ``float`` of the percentage of at bats that result in the
        batter getting on base.
        """
        return self._on_base_percentage

    @_float_property_decorator
    def slugging_percentage(self):
        """
        Returns a ``float`` of the slugging percentage for the player based
        on the number of bases gained per at-bat with bigger plays getting more
        weight.
        """
        return self._slugging_percentage

    @_float_property_decorator
    def on_base_plus_slugging_percentage(self):
        """
        Returns a ``float`` of the on base percentage plus the slugging
        percentage. Percentage ranges from 0-1.
        """
        return self._on_base_plus_slugging_percentage

    @_int_property_decorator
    def putouts(self):
        """
        Returns an ``int`` of the number of putouts the player had.
        """
        return self._putouts

    @_int_property_decorator
    def assists(self):
        """
        Returns an ``int`` of the number of assists the player had.
        """
        return self._assists

    @_int_property_decorator
    def hits_allowed(self):
        """
        Returns an ``int`` of the number of hits the player allowed as a
        pitcher.
        """
        return self._hits_allowed

    @_int_property_decorator
    def runs_allowed(self):
        """
        Returns an ``int`` of the number of runs the player allowed as a
        pitcher.
        """
        return self._runs_allowed

    @_int_property_decorator
    def earned_runs_allowed(self):
        """
        Returns an ``int`` of the number of earned runs the player allowed as a
        pitcher.
        """
        return self._earned_runs_allowed

    @_int_property_decorator
    def bases_on_balls_given(self):
        """
        Returns an ``int`` of the number of bases on balls the player has given
        as a pitcher.
        """
        return self._bases_on_balls_given

    @_int_property_decorator
    def strikeouts(self):
        """
        Returns an ``int`` of the number of strikeouts the player threw as a
        pitcher.
        """
        return self._strikeouts

    @_int_property_decorator
    def batters_faced(self):
        """
        Returns an ``int`` of the number of batters the pitcher has faced.
        """
        return self._batters_faced



from flexmock import flexmock
from mock import PropertyMock
from sportsreference.constants import (AWAY,
                                       HOME,
                                       LOSS,
                                       NEUTRAL,
                                       POST_SEASON,
                                       REGULAR_SEASON,
                                       WIN)
from sportsreference.nhl.constants import OVERTIME_LOSS, SHOOTOUT
from sportsreference.nhl.schedule import Game, Schedule


YEAR = 2017


class TestNHLSchedule:
    def setup_method(self, *args, **kwargs):
        flexmock(Game) \
            .should_receive('_parse_game_data') \
            .and_return(None)

        self.game = Game(None, YEAR)

    def test_away_game_returns_away_location(self):
        fake_location = PropertyMock(return_value='@')
        type(self.game)._location = fake_location

        assert self.game.location == AWAY

    def test_home_game_returns_home_location(self):
        fake_location = PropertyMock(return_value='')
        type(self.game)._location = fake_location

        assert self.game.location == HOME

    def test_winning_result_returns_win(self):
        fake_result = PropertyMock(return_value='W')
        type(self.game)._result = fake_result

        assert self.game.result == WIN

    def test_losing_result_in_overtime_returns_overtime_loss(self):
        fake_result = PropertyMock(return_value='L')
        type(self.game)._result = fake_result
        fake_overtime = PropertyMock(return_value='OT')
        type(self.game)._overtime = fake_overtime

        assert self.game.result == OVERTIME_LOSS

    def test_losing_result_in_regulation_returns_loss(self):
        fake_result = PropertyMock(return_value='L')
        type(self.game)._result = fake_result
        fake_overtime = PropertyMock(return_value='')
        type(self.game)._overtime = fake_overtime

        assert self.game.result == LOSS

    def test_overtime_returns_one(self):
        fake_overtime = PropertyMock(return_value='OT')
        type(self.game)._overtime = fake_overtime

        assert self.game.overtime == 1

    def test_no_overtime_returns_zero(self):
        fake_overtime = PropertyMock(return_value='')
        type(self.game)._overtime = fake_overtime

        assert self.game.overtime == 0

    def test_double_overtime_returns_two(self):
        fake_overtime = PropertyMock(return_value='2OT')
        type(self.game)._overtime = fake_overtime

        assert self.game.overtime == 2

    def test_shootout_returns_shootout_constant(self):
        fake_overtime = PropertyMock(return_value='SO')
        type(self.game)._overtime = fake_overtime

        assert self.game.overtime == SHOOTOUT

    def test_bad_overtime_returns_default_number(self):
        fake_overtime = PropertyMock(return_value='BAD')
        type(self.game)._overtime = fake_overtime

        assert self.game.overtime == 0

    def test_bad_shots_on_goal_returns_default_number(self):
        fake_shots = PropertyMock(return_value='')
        type(self.game)._shots_on_goal = fake_shots

        assert self.game.shots_on_goal is None

    def test_bad_penalties_in_minutes_returns_default_number(self):
        fake_penalties = PropertyMock(return_value='')
        type(self.game)._penalties_in_minutes = fake_penalties

        assert self.game.penalties_in_minutes is None

    def test_bad_power_play_goals_returns_default_number(self):
        fake_power_play_goals = PropertyMock(return_value='')
        type(self.game)._power_play_goals = fake_power_play_goals

        assert self.game.power_play_goals is None

    def test_bad_power_play_opportunities_returns_default_number(self):
        fake_power_play_opps = PropertyMock(return_value='')
        type(self.game)._power_play_opportunities = fake_power_play_opps

        assert self.game.power_play_opportunities is None

    def test_bad_short_handed_goals_returns_default_number(self):
        fake_goals = PropertyMock(return_value='')
        type(self.game)._short_handed_goals = fake_goals

        assert self.game.short_handed_goals is None

    def test_bad_opp_shots_on_goal_returns_default_number(self):
        fake_shots = PropertyMock(return_value='')
        type(self.game)._opp_shots_on_goal = fake_shots

        assert self.game.opp_shots_on_goal is None

    def test_bad_opp_penalties_in_minutes_returns_default_number(self):
        fake_penalties = PropertyMock(return_value='')
        type(self.game)._opp_penalties_in_minutes = fake_penalties

        assert self.game.opp_penalties_in_minutes is None

    def test_bad_opp_power_play_goals_returns_default_number(self):
        fake_power_play_goals = PropertyMock(return_value='')
        type(self.game)._opp_power_play_goals = fake_power_play_goals

        assert self.game.opp_power_play_goals is None

    def test_bad_opp_power_play_opportunities_returns_default_number(self):
        fake_power_play_opps = PropertyMock(return_value='')
        type(self.game)._opp_power_play_opportunities = fake_power_play_opps

        assert self.game.opp_power_play_opportunities is None

    def test_bad_opp_short_handed_goals_returns_default_number(self):
        fake_goals = PropertyMock(return_value='')
        type(self.game)._opp_short_handed_goals = fake_goals

        assert self.game.opp_short_handed_goals is None

    def test_bad_penalties_in_minutes_returns_default_number(self):
        fake_penalties = PropertyMock(return_value='')
        type(self.game)._penalties_in_minutes = fake_penalties

        assert self.game.penalties_in_minutes is None

    def test_bad_power_play_goals_returns_default_number(self):
        fake_power_play_goals = PropertyMock(return_value='')
        type(self.game)._power_play_goals = fake_power_play_goals

        assert self.game.power_play_goals is None

    def test_bad_power_play_opportunities_returns_default_number(self):
        fake_power_play_opps = PropertyMock(return_value='')
        type(self.game)._power_play_opportunities = fake_power_play_opps

        assert self.game.power_play_opportunities is None

    def test_bad_corsi_for_returns_default_number(self):
        fake_corsi = PropertyMock(return_value='')
        type(self.game)._corsi_for = fake_corsi

        assert self.game.corsi_for is None

    def test_bad_corsi_against_returns_default_number(self):
        fake_corsi = PropertyMock(return_value='')
        type(self.game)._corsi_against = fake_corsi

        assert self.game.corsi_against is None

    def test_bad_corsi_for_percentage_returns_default_number(self):
        fake_corsi = PropertyMock(return_value='')
        type(self.game)._corsi_for_percentage = fake_corsi

        assert self.game.corsi_for_percentage is None

    def test_bad_fenwick_for_returns_default_number(self):
        fake_fenwick = PropertyMock(return_value='')
        type(self.game)._fenwick_for = fake_fenwick

        assert self.game.fenwick_for is None

    def test_bad_fenwick_against_returns_default_number(self):
        fake_fenwick = PropertyMock(return_value='')
        type(self.game)._fenwick_against = fake_fenwick

        assert self.game.fenwick_against is None

    def test_bad_fenwick_for_percentage_returns_default_number(self):
        fake_fenwick = PropertyMock(return_value='')
        type(self.game)._fenwick_for_percentage = fake_fenwick

        assert self.game.fenwick_for_percentage is None

    def test_bad_faceoff_wins_returns_default_number(self):
        fake_faceoff = PropertyMock(return_value='')
        type(self.game)._faceoff_wins = fake_faceoff

        assert self.game.faceoff_wins is None

    def test_bad_faceoff_losses_returns_default_number(self):
        fake_faceoff = PropertyMock(return_value='')
        type(self.game)._faceoff_losses = fake_faceoff

        assert self.game.faceoff_losses is None

    def test_bad_faceoff_win_percentage_returns_default_number(self):
        fake_faceoff = PropertyMock(return_value='')
        type(self.game)._faceoff_win_percentage = fake_faceoff

        assert self.game.faceoff_win_percentage is None

    def test_bad_offensive_zone_start_percentage_returns_default_number(self):
        fake_start = PropertyMock(return_value='')
        type(self.game)._offensive_zone_start_percentage = fake_start

        assert self.game.offensive_zone_start_percentage is None

    def test_bad_pdo_returns_default_number(self):
        fake_pdo = PropertyMock(return_value='')
        type(self.game)._pdo = fake_pdo

        assert self.game.pdo is None

    def test_empty_game_class_returns_dataframe_of_none(self):
        assert self.game._goals_scored is None
        assert self.game._goals_allowed is None
        assert self.game.dataframe is None

    def test_no_dataframes_returns_none(self):
        flexmock(Schedule) \
            .should_receive('_pull_schedule') \
            .and_return(None)
        schedule = Schedule('DET')

        fake_game = flexmock(dataframe=None)
        fake_games = PropertyMock(return_value=fake_game)
        type(schedule).__iter__ = fake_games

        assert schedule.dataframe is None

    def test_no_dataframes_extended_returns_none(self):
        flexmock(Schedule) \
            .should_receive('_pull_schedule') \
            .and_return(None)
        schedule = Schedule('DET')

        fake_game = flexmock(dataframe_extended=None)
        fake_games = PropertyMock(return_value=fake_game)
        type(schedule).__iter__ = fake_games

        assert schedule.dataframe_extended is None



import mock
import os
import pandas as pd
import pytest
from flexmock import flexmock
from sportsreference import utils
from sportsreference.nhl.roster import Player, Roster
from sportsreference.nhl.teams import Team


def read_file(filename):
    filepath = os.path.join(os.path.dirname(__file__), 'nhl', filename)
    return open('%s.html' % filepath, 'r').read()


def mock_pyquery(url):
    class MockPQ:
        def __init__(self, html_contents, status=200):
            self.url = url
            self.reason = 'Bad URL'  # Used when throwing HTTPErrors
            self.headers = {}  # Used when throwing HTTPErrors
            self.status_code = status
            self.html_contents = html_contents
            self.text = html_contents

    if 'BAD' in url or 'bad' in url:
        return MockPQ(None, 404)
    if 'zettehe01' in url:
        return MockPQ(read_file('zettehe01'))
    if '2018' in url:
        return MockPQ(read_file('2018'))
    return MockPQ(read_file('howarja02'))


class TestNHLPlayer:
    def setup_method(self):
        self.skater_results_career = {
            'adjusted_assists': 692,
            'adjusted_goals': 377,
            'adjusted_goals_against_average': None,
            'adjusted_goals_created': 394,
            'adjusted_points': 1069,
            'age': None,
            'assists': 623,
            'average_time_on_ice': '19:35',
            'blocks_at_even_strength': 267,
            'corsi_against': 10322.0,
            'corsi_for': 12688,
            'corsi_for_percentage': 55.1,
            'defensive_point_shares': 29.4,
            'defensive_zone_start_percentage': 45.5,
            'even_strength_assists': 379,
            'even_strength_goals': 228,
            'even_strength_goals_allowed': None,
            'even_strength_save_percentage': None,
            'even_strength_shots_faced': None,
            'faceoff_losses': 5602,
            'faceoff_percentage': 51.1,
            'faceoff_wins': 5863,
            'fenwick_against': 8123,
            'fenwick_for': 9757,
            'fenwick_for_percentage': 54.6,
            'game_winning_goals': 64,
            'games_played': 1082,
            'giveaways': 482,
            'goal_against_percentage_relative': None,
            'goalie_point_shares': None,
            'goals': 337,
            'goals_against': None,
            'goals_against_average': None,
            'goals_against_on_ice': 530,
            'goals_created': 348,
            'goals_for_on_ice': 633,
            'goals_saved_above_average': None,
            'height': '6-0',
            'hits_at_even_strength': 471,
            'league': 'NHL',
            'losses': None,
            'minutes': None,
            'name': 'Henrik Zetterberg',
            'offensive_point_shares': 79.9,
            'offensive_zone_start_percentage': 54.5,
            'pdo': 100.0,
            'penalties_in_minutes': 401,
            'player_id': 'zettehe01',
            'plus_minus': 160,
            'point_shares': 109.3,
            'points': 960,
            'power_play_assists': 235,
            'power_play_goals': 100,
            'power_play_goals_against_on_ice': 140,
            'power_play_goals_allowed': None,
            'power_play_goals_for_on_ice': 490,
            'power_play_save_percentage': None,
            'power_play_shots_faced': None,
            'quality_start_percentage': None,
            'quality_starts': None,
            'really_bad_starts': None,
            'relative_corsi_for_percentage': 3.3,
            'relative_fenwick_for_percentage': 3.1,
            'save_percentage': None,
            'save_percentage_on_ice': None,
            'saves': None,
            'season': 'Career',
            'shooting_percentage': 9.8,
            'shooting_percentage_on_ice': 8.8,
            'shootout_attempts': 47,
            'shootout_goals': 10,
            'shootout_misses': 37,
            'shootout_percentage': 21.3,
            'short_handed_assists': 9,
            'short_handed_goals': 9,
            'short_handed_goals_allowed': None,
            'short_handed_save_percentage': None,
            'short_handed_shots_faced': None,
            'shots_against': None,
            'shots_on_goal': 3455,
            'shutouts': None,
            'takeaways': 454,
            'team_abbreviation': None,
            'ties_plus_overtime_loss': None,
            'time_on_ice': 21186,
            'time_on_ice_even_strength': 12658.7,
            'total_goals_against_on_ice': 851,
            'total_goals_for_on_ice': 1362,
            'total_shots': 5408,
            'weight': 197,
            'wins': None
        }

        self.skater_results_2017 = {
            'adjusted_assists': 46,
            'adjusted_goals': 11,
            'adjusted_goals_against_average': None,
            'adjusted_goals_created': 19,
            'adjusted_points': 57,
            'age': 37,
            'assists': 45,
            'average_time_on_ice': '19:30',
            'blocks_at_even_strength': 34,
            'corsi_against': 1243.0,
            'corsi_for': 1274,
            'corsi_for_percentage': 50.6,
            'defensive_point_shares': 2.0,
            'defensive_zone_start_percentage': 45.2,
            'even_strength_assists': 28,
            'even_strength_goals': 10,
            'even_strength_goals_allowed': None,
            'even_strength_save_percentage': None,
            'even_strength_shots_faced': None,
            'faceoff_losses': 709,
            'faceoff_percentage': 48.4,
            'faceoff_wins': 666,
            'fenwick_against': 948,
            'fenwick_for': 975,
            'fenwick_for_percentage': 50.7,
            'game_winning_goals': 2,
            'games_played': 82,
            'giveaways': 57,
            'goal_against_percentage_relative': None,
            'goalie_point_shares': None,
            'goals': 11,
            'goals_against': None,
            'goals_against_average': None,
            'goals_against_on_ice': 52,
            'goals_created': 18,
            'goals_for_on_ice': 54,
            'goals_saved_above_average': None,
            'height': '6-0',
            'hits_at_even_strength': 49,
            'league': 'NHL',
            'losses': None,
            'minutes': None,
            'name': 'Henrik Zetterberg',
            'offensive_point_shares': 2.4,
            'offensive_zone_start_percentage': 54.8,
            'pdo': 99.9,
            'penalties_in_minutes': 14,
            'player_id': 'zettehe01',
            'plus_minus': 1,
            'point_shares': 4.4,
            'points': 56,
            'power_play_assists': 17,
            'power_play_goals': 1,
            'power_play_goals_against_on_ice': 0,
            'power_play_goals_allowed': None,
            'power_play_goals_for_on_ice': 25,
            'power_play_save_percentage': None,
            'power_play_shots_faced': None,
            'quality_start_percentage': None,
            'quality_starts': None,
            'really_bad_starts': None,
            'relative_corsi_for_percentage': 2.7,
            'relative_fenwick_for_percentage': 2.0,
            'save_percentage': None,
            'save_percentage_on_ice': None,
            'saves': None,
            'season': '2017-18',
            'shooting_percentage': 6.1,
            'shooting_percentage_on_ice': 7.6,
            'shootout_attempts': 3,
            'shootout_goals': 0,
            'shootout_misses': 3,
            'shootout_percentage': 0.0,
            'short_handed_assists': 0,
            'short_handed_goals': 0,
            'short_handed_goals_allowed': None,
            'short_handed_save_percentage': None,
            'short_handed_shots_faced': None,
            'shots_against': None,
            'shots_on_goal': 180,
            'shutouts': None,
            'takeaways': 51,
            'team_abbreviation': 'DET',
            'ties_plus_overtime_loss': None,
            'time_on_ice': 1599,
            'time_on_ice_even_strength': 1382.2,
            'total_goals_against_on_ice': 53,
            'total_goals_for_on_ice': 79,
            'total_shots': 332,
            'weight': 197,
            'wins': None
        }

        self.goalie_results_career = {
            'adjusted_assists': None,
            'adjusted_goals': None,
            'adjusted_goals_against_average': None,
            'adjusted_goals_created': None,
            'adjusted_points': None,
            'age': None,
            'assists': 8,
            'average_time_on_ice': None,
            'blocks_at_even_strength': None,
            'corsi_against': None,
            'corsi_for': None,
            'corsi_for_percentage': None,
            'defensive_point_shares': None,
            'defensive_zone_start_percentage': None,
            'even_strength_assists': None,
            'even_strength_goals': None,
            'even_strength_goals_allowed': 800,
            'even_strength_save_percentage': 0.922,
            'even_strength_shots_faced': 10295,
            'faceoff_losses': None,
            'faceoff_percentage': None,
            'faceoff_wins': None,
            'fenwick_against': None,
            'fenwick_for': None,
            'fenwick_for_percentage': None,
            'game_winning_goals': None,
            'games_played': None,
            'giveaways': None,
            'goal_against_percentage_relative': 97,
            'goalie_point_shares': 78.8,
            'goals': 0,
            'goals_against': 1091,
            'goals_against_average': 2.49,
            'goals_against_on_ice': None,
            'goals_created': None,
            'goals_for_on_ice': None,
            'goals_saved_above_average': None,
            'height': '6-1',
            'hits_at_even_strength': None,
            'league': 'NHL',
            'losses': 151,
            'minutes': 26332,
            'name': 'Jimmy Howard',
            'offensive_point_shares': None,
            'offensive_zone_start_percentage': None,
            'pdo': None,
            'penalties_in_minutes': 34,
            'player_id': 'howarja02',
            'plus_minus': None,
            'point_shares': None,
            'points': 8,
            'power_play_assists': None,
            'power_play_goals': None,
            'power_play_goals_against_on_ice': None,
            'power_play_goals_allowed': 26,
            'power_play_goals_for_on_ice': None,
            'power_play_save_percentage': 0.92,
            'power_play_shots_faced': 327,
            'quality_start_percentage': 0.544,
            'quality_starts': 239,
            'really_bad_starts': 61,
            'relative_corsi_for_percentage': None,
            'relative_fenwick_for_percentage': None,
            'save_percentage': 0.915,
            'save_percentage_on_ice': None,
            'saves': 11696,
            'season': 'Career',
            'shooting_percentage': None,
            'shooting_percentage_on_ice': None,
            'shootout_attempts': None,
            'shootout_goals': None,
            'shootout_misses': None,
            'shootout_percentage': None,
            'short_handed_assists': None,
            'short_handed_goals': None,
            'short_handed_goals_allowed': 249,
            'short_handed_save_percentage': 0.877,
            'short_handed_shots_faced': 2027,
            'shots_against': 12787,
            'shots_on_goal': None,
            'shutouts': 24,
            'takeaways': None,
            'team_abbreviation': None,
            'ties_plus_overtime_loss': 63,
            'time_on_ice': None,
            'time_on_ice_even_strength': None,
            'total_goals_against_on_ice': None,
            'total_goals_for_on_ice': None,
            'total_shots': None,
            'weight': 218,
            'wins': 221
        }

        self.goalie_results_2017 = {
            'adjusted_assists': None,
            'adjusted_goals': None,
            'adjusted_goals_against_average': None,
            'adjusted_goals_created': None,
            'adjusted_points': None,
            'age': 33,
            'assists': 1,
            'average_time_on_ice': None,
            'blocks_at_even_strength': None,
            'corsi_against': None,
            'corsi_for': None,
            'corsi_for_percentage': None,
            'defensive_point_shares': None,
            'defensive_zone_start_percentage': None,
            'even_strength_assists': None,
            'even_strength_goals': None,
            'even_strength_goals_allowed': 122,
            'even_strength_save_percentage': 0.916,
            'even_strength_shots_faced': 1455,
            'faceoff_losses': None,
            'faceoff_percentage': None,
            'faceoff_wins': None,
            'fenwick_against': None,
            'fenwick_for': None,
            'fenwick_for_percentage': None,
            'game_winning_goals': None,
            'games_played': None,
            'giveaways': None,
            'goal_against_percentage_relative': 103,
            'goalie_point_shares': 9.4,
            'goals': 0,
            'goals_against': 160,
            'goals_against_average': 2.85,
            'goals_against_on_ice': None,
            'goals_created': None,
            'goals_for_on_ice': None,
            'goals_saved_above_average': -4.65,
            'height': '6-1',
            'hits_at_even_strength': None,
            'league': 'NHL',
            'losses': 27,
            'minutes': 3368,
            'name': 'Jimmy Howard',
            'offensive_point_shares': None,
            'offensive_zone_start_percentage': None,
            'pdo': None,
            'penalties_in_minutes': 10,
            'player_id': 'howarja02',
            'plus_minus': None,
            'point_shares': None,
            'points': 1,
            'power_play_assists': None,
            'power_play_goals': None,
            'power_play_goals_against_on_ice': None,
            'power_play_goals_allowed': 2,
            'power_play_goals_for_on_ice': None,
            'power_play_save_percentage': 0.949,
            'power_play_shots_faced': 39,
            'quality_start_percentage': 0.491,
            'quality_starts': 28,
            'really_bad_starts': 6,
            'relative_corsi_for_percentage': None,
            'relative_fenwick_for_percentage': None,
            'save_percentage': 0.91,
            'save_percentage_on_ice': None,
            'saves': 1610,
            'season': '2017-18',
            'shooting_percentage': None,
            'shooting_percentage_on_ice': None,
            'shootout_attempts': None,
            'shootout_goals': None,
            'shootout_misses': None,
            'shootout_percentage': None,
            'short_handed_assists': None,
            'short_handed_goals': None,
            'short_handed_goals_allowed': 36,
            'short_handed_save_percentage': 0.869,
            'short_handed_shots_faced': 275,
            'shots_against': 1770,
            'shots_on_goal': None,
            'shutouts': 1,
            'takeaways': None,
            'team_abbreviation': 'DET',
            'ties_plus_overtime_loss': 9,
            'time_on_ice': None,
            'time_on_ice_even_strength': None,
            'total_goals_against_on_ice': None,
            'total_goals_for_on_ice': None,
            'total_shots': None,
            'weight': 218,
            'wins': 22
        }

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_nhl_skater_returns_requested_career_stats(self, *args, **kwargs):
        # Request the career stats
        player = Player('zettehe01')
        player = player('')

        for attribute, value in self.skater_results_career.items():
            assert getattr(player, attribute) == value

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_nhl_skater_returns_player_season_stats(self, *args, **kwargs):
        # Request the 2017 stats
        player = Player('zettehe01')
        player = player('2017-18')

        for attribute, value in self.skater_results_2017.items():
            assert getattr(player, attribute) == value

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_nhl_goalie_returns_requested_career_stats(self, *args, **kwargs):
        # Request the career stats
        player = Player('howarja02')
        player = player('')

        for attribute, value in self.goalie_results_career.items():
            assert getattr(player, attribute) == value

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_nhl_goalie_returns_player_season_stats(self, *args, **kwargs):
        # Request the 2017 stats
        player = Player('howarja02')
        player = player('2017-18')

        for attribute, value in self.goalie_results_2017.items():
            assert getattr(player, attribute) == value

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_dataframe_returns_dataframe(self, *args, **kwargs):
        dataframe = [
            {'adjusted_assists': 46,
             'adjusted_goals': 11,
             'adjusted_goals_against_average': None,
             'adjusted_goals_created': 19,
             'adjusted_points': 57,
             'age': 37,
             'assists': 45,
             'average_time_on_ice': '19:30',
             'blocks_at_even_strength': 34,
             'corsi_against': 1243.0,
             'corsi_for': None,
             'corsi_for_percentage': 50.6,
             'defensive_point_shares': 2.0,
             'defensive_zone_start_percentage': 45.2,
             'even_strength_assists': 28,
             'even_strength_goals': 10,
             'even_strength_goals_allowed': None,
             'even_strength_save_percentage': None,
             'even_strength_shots_faced': None,
             'faceoff_losses': 709,
             'faceoff_percentage': 48.4,
             'faceoff_wins': 666,
             'fenwick_against': 948,
             'fenwick_for': 975,
             'fenwick_for_percentage': 50.7,
             'game_winning_goals': 2,
             'games_played': 82,
             'giveaways': 57,
             'goal_against_percentage_relative': None,
             'goalie_point_shares': None,
             'goals': 11,
             'goals_against': None,
             'goals_against_average': None,
             'goals_against_on_ice': 52,
             'goals_created': 18,
             'goals_for_on_ice': 54,
             'goals_saved_above_average': None,
             'height': '6-0',
             'hits_at_even_strength': 49,
             'league': 'NHL',
             'losses': None,
             'minutes': None,
             'name': 'Henrik Zetterberg',
             'offensive_point_shares': 2.4,
             'offensive_zone_start_percentage': 54.8,
             'pdo': 99.9,
             'penalties_in_minutes': 14,
             'player_id': 'zettehe01',
             'plus_minus': 1,
             'point_shares': 4.4,
             'points': 56,
             'power_play_assists': 17,
             'power_play_goals': 1,
             'power_play_goals_against_on_ice': 0,
             'power_play_goals_allowed': None,
             'power_play_goals_for_on_ice': 25,
             'power_play_save_percentage': None,
             'power_play_shots_faced': None,
             'quality_start_percentage': None,
             'quality_starts': None,
             'really_bad_starts': None,
             'relative_corsi_for_percentage': 2.7,
             'relative_fenwick_for_percentage': 2.0,
             'save_percentage': None,
             'save_percentage_on_ice': None,
             'saves': None,
             'season': '2017-18',
             'shooting_percentage': 6.1,
             'shooting_percentage_on_ice': 7.6,
             'shootout_attempts': 3,
             'shootout_goals': 0,
             'shootout_misses': 3,
             'shootout_percentage': 0.0,
             'short_handed_assists': 0,
             'short_handed_goals': 0,
             'short_handed_goals_allowed': None,
             'short_handed_save_percentage': None,
             'short_handed_shots_faced': None,
             'shots_against': None,
             'shots_on_goal': 180,
             'shutouts': None,
             'takeaways': 51,
             'team_abbreviation': 'DET',
             'ties_plus_overtime_loss': None,
             'time_on_ice': 1599,
             'time_on_ice_even_strength': 1382.2,
             'total_goals_against_on_ice': 53,
             'total_goals_for_on_ice': 79,
             'total_shots': 332,
             'weight': 197,
             'wins': None},
            {'adjusted_assists': 692,
             'adjusted_goals': 377,
             'adjusted_goals_against_average': None,
             'adjusted_goals_created': 394,
             'adjusted_points': 1069,
             'age': None,
             'assists': 623,
             'average_time_on_ice': '19:35',
             'blocks_at_even_strength': 267,
             'corsi_against': 10322.0,
             'corsi_for': None,
             'corsi_for_percentage': 55.1,
             'defensive_point_shares': 29.4,
             'defensive_zone_start_percentage': 45.5,
             'even_strength_assists': 379,
             'even_strength_goals': 228,
             'even_strength_goals_allowed': None,
             'even_strength_save_percentage': None,
             'even_strength_shots_faced': None,
             'faceoff_losses': 5602,
             'faceoff_percentage': 51.1,
             'faceoff_wins': 5863,
             'fenwick_against': 8123,
             'fenwick_for': 9757,
             'fenwick_for_percentage': 54.6,
             'game_winning_goals': 64,
             'games_played': 1082,
             'giveaways': 482,
             'goal_against_percentage_relative': None,
             'goalie_point_shares': None,
             'goals': 337,
             'goals_against': None,
             'goals_against_average': None,
             'goals_against_on_ice': 530,
             'goals_created': 348,
             'goals_for_on_ice': 633,
             'goals_saved_above_average': None,
             'height': '6-0',
             'hits_at_even_strength': 471,
             'league': 'NHL',
             'losses': None,
             'minutes': None,
             'name': 'Henrik Zetterberg',
             'offensive_point_shares': 79.9,
             'offensive_zone_start_percentage': 54.5,
             'pdo': 100.0,
             'penalties_in_minutes': 401,
             'player_id': 'zettehe01',
             'plus_minus': 160,
             'point_shares': 109.3,
             'points': 960,
             'power_play_assists': 235,
             'power_play_goals': 100,
             'power_play_goals_against_on_ice': 140,
             'power_play_goals_allowed': None,
             'power_play_goals_for_on_ice': 490,
             'power_play_save_percentage': None,
             'power_play_shots_faced': None,
             'quality_start_percentage': None,
             'quality_starts': None,
             'really_bad_starts': None,
             'relative_corsi_for_percentage': 3.3,
             'relative_fenwick_for_percentage': 3.1,
             'save_percentage': None,
             'save_percentage_on_ice': None,
             'saves': None,
             'season': 'Career',
             'shooting_percentage': 9.8,
             'shooting_percentage_on_ice': 8.8,
             'shootout_attempts': 47,
             'shootout_goals': 10,
             'shootout_misses': 37,
             'shootout_percentage': 21.3,
             'short_handed_assists': 9,
             'short_handed_goals': 9,
             'short_handed_goals_allowed': None,
             'short_handed_save_percentage': None,
             'short_handed_shots_faced': None,
             'shots_against': None,
             'shots_on_goal': 3455,
             'shutouts': None,
             'takeaways': 454,
             'team_abbreviation': None,
             'ties_plus_overtime_loss': None,
             'time_on_ice': 21186,
             'time_on_ice_even_strength': 12658.7,
             'total_goals_against_on_ice': 851,
             'total_goals_for_on_ice': 1362,
             'total_shots': 5408,
             'weight': 197,
             'wins': None}
        ]
        indices = ['2017', 'Career']

        df = pd.DataFrame(dataframe, index=indices)
        player = Player('zettehe01')

        # Pandas doesn't natively allow comparisons of DataFrames.
        # Concatenating the two DataFrames (the one generated during the test
        # and the expected on above) and dropping duplicate rows leaves only
        # the rows that are unique between the two frames. This allows a quick
        # check of the DataFrame to see if it is empty - if so, all rows are
        # duplicates, and they are equal.
        frames = [df, player.dataframe]
        df1 = pd.concat(frames).drop_duplicates(keep=False)

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_nhl_404_returns_none_with_no_errors(self, *args, **kwargs):
        player = Player('bad')

        assert player.name is None
        assert player.dataframe is None

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_nhl_404_returns_none_for_different_season(self, *args, **kwargs):
        player = Player('bad')

        assert player.name is None
        assert player.dataframe is None


class TestNHLRoster:
    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_roster_class_pulls_all_player_stats(self, *args, **kwargs):
        flexmock(utils) \
            .should_receive('_find_year_for_season') \
            .and_return('2018')
        roster = Roster('DET')

        assert len(roster.players) == 2

        for player in roster.players:
            assert player.name in ['Jimmy Howard', 'Henrik Zetterberg']

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_bad_url_raises_value_error(self, *args, **kwargs):
        with pytest.raises(ValueError):
            roster = Roster('bad')

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_roster_from_team_class(self, *args, **kwargs):
        flexmock(Team) \
            .should_receive('_parse_team_data') \
            .and_return(None)
        team = Team(None, 1, '2018')
        mock_abbreviation = mock.PropertyMock(return_value='DET')
        type(team)._abbreviation = mock_abbreviation

        assert len(team.roster.players) == 2

        for player in team.roster.players:
            assert player.name in ['Jimmy Howard', 'Henrik Zetterberg']
        type(team)._abbreviation = None

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_roster_class_with_slim_parameter(self, *args, **kwargs):
        flexmock(utils) \
            .should_receive('_find_year_for_season') \
            .and_return('2018')
        roster = Roster('DET', slim=True)

        assert len(roster.players) == 2
        assert roster.players == {
            'howarja02': 'Jimmy Howard',
            'zettehe01': 'Henrik Zetterberg'
        }



# coding=utf-8
import mock
import os
import pandas as pd
import pytest
from flexmock import flexmock
from sportsreference import utils
from sportsreference.mlb.roster import Player, Roster
from sportsreference.mlb.teams import Team


def read_file(filename):
    filepath = os.path.join(os.path.dirname(__file__), 'mlb', filename)
    return open('%s.shtml' % filepath, 'r').read()


def mock_pyquery(url):
    class MockPQ:
        def __init__(self, html_contents, status=200):
            self.url = url
            self.reason = 'Bad URL'  # Used when throwing HTTPErrors
            self.headers = {}  # Used when throwing HTTPErrors
            self.status_code = status
            self.html_contents = html_contents
            self.text = html_contents

    if 'BAD' in url:
        return MockPQ(None, 404)
    if 'HOU' in url:
        return MockPQ(read_file('2017'))
    if 'verlaju01' in url:
        return MockPQ(read_file('verlaju01'))
    return MockPQ(read_file('altuvjo01'))


class TestMLBPlayer:
    @mock.patch('requests.get', side_effect=mock_pyquery)
    def setup_method(self, *args, **kwargs):
        self.results_career = {
            'assists': 2763,
            'at_bats': 4436,
            'bases_on_balls': 311,
            'batting_average': 0.317,
            'birth_date': '1990-05-06',
            'complete_games': 990,
            'defensive_chances': 4500,
            'defensive_runs_saved_above_average': -24,
            'defensive_runs_saved_above_average_per_innings': -3,
            'double_plays_turned': 623,
            'doubles': 271,
            'errors': 61,
            'fielding_percentage': 0.986,
            'games': 1105,
            'games_catcher': 0,
            'games_center_fielder': 0,
            'games_designated_hitter': 30,
            'games_first_baseman': 0,
            'games_in_batting_order': 1105,
            'games_in_defensive_lineup': 1070,
            'games_left_fielder': 0,
            'games_outfielder': 0,
            'games_pinch_hitter': 7,
            'games_pinch_runner': 4,
            'games_pitcher': 0,
            'games_right_fielder': 0,
            'games_second_baseman': 1070,
            'games_shortstop': 1,
            'games_started': 1059,
            'games_third_baseman': 0,
            'grounded_into_double_plays': 123,
            'height': '5-6',
            'hits': 1404,
            'home_runs': 96,
            'innings_played': 9269.0,
            'intentional_bases_on_balls': 38,
            'league_fielding_percentage': 0.984,
            'league_range_factor_per_game': 4.5,
            'league_range_factor_per_nine_innings': 4.53,
            'name': u'José Altuve',
            'nationality': 'Venezuela',
            'on_base_percentage': 0.364,
            'on_base_plus_slugging_percentage': 0.818,
            'on_base_plus_slugging_percentage_plus': 126,
            'plate_appearances': 4851,
            'player_id': 'altuvjo01',
            'position': '2B',
            'putouts': 1676,
            'range_factor_per_game': 4.15,
            'range_factor_per_nine_innings': 4.31,
            'runs': 635,
            'runs_batted_in': 457,
            'sacrifice_flies': 36,
            'sacrifice_hits': 24,
            'season': 'Career',
            'slugging_percentage': 0.454,
            'stolen_bases': 246,
            'team_abbreviation': 'HOU',
            'times_caught_stealing': 68,
            'times_hit_by_pitch': 44,
            'times_struck_out': 535,
            'total_bases': 2013,
            'total_fielding_runs_above_average': 23,
            'total_fielding_runs_above_average_per_innings': 3,
            'triples': 25,
            'weight': 165
        }

        self.results_2017 = {
            'assists': 351,
            'at_bats': 590,
            'bases_on_balls': 58,
            'batting_average': 0.346,
            'birth_date': '1990-05-06',
            'complete_games': 134,
            'defensive_chances': 562,
            'defensive_runs_saved_above_average': 3,
            'defensive_runs_saved_above_average_per_innings': 3,
            'double_plays_turned': 86,
            'doubles': 39,
            'errors': 10,
            'fielding_percentage': 0.982,
            'games': 153,
            'games_catcher': 0,
            'games_center_fielder': 0,
            'games_designated_hitter': 3,
            'games_first_baseman': 0,
            'games_in_batting_order': 153,
            'games_in_defensive_lineup': 149,
            'games_left_fielder': 0,
            'games_outfielder': 0,
            'games_pinch_hitter': 1,
            'games_pinch_runner': 1,
            'games_pitcher': 0,
            'games_right_fielder': 0,
            'games_second_baseman': 149,
            'games_shortstop': 0,
            'games_started': 148,
            'games_third_baseman': 0,
            'grounded_into_double_plays': 19,
            'height': '5-6',
            'hits': 204,
            'home_runs': 24,
            'innings_played': 1283.1,
            'intentional_bases_on_balls': 3,
            'league_fielding_percentage': 0.983,
            'league_range_factor_per_game': 4.3,
            'league_range_factor_per_nine_innings': 4.35,
            'name': u'José Altuve',
            'nationality': 'Venezuela',
            'on_base_percentage': 0.41,
            'on_base_plus_slugging_percentage': 0.957,
            'on_base_plus_slugging_percentage_plus': 162,
            'plate_appearances': 662,
            'player_id': 'altuvjo01',
            'position': '2B',
            'putouts': 201,
            'range_factor_per_game': 3.7,
            'range_factor_per_nine_innings': 3.87,
            'runs': 112,
            'runs_batted_in': 81,
            'sacrifice_flies': 4,
            'sacrifice_hits': 1,
            'season': '2017',
            'slugging_percentage': 0.547,
            'stolen_bases': 32,
            'team_abbreviation': 'HOU',
            'times_caught_stealing': 6,
            'times_hit_by_pitch': 9,
            'times_struck_out': 84,
            'total_bases': 323,
            'total_fielding_runs_above_average': 13,
            'total_fielding_runs_above_average_per_innings': 12,
            'triples': 4,
            'weight': 165,
            # Properties specific to pitchers
            'balks': None,
            'bases_on_balls_given': None,
            'bases_on_balls_given_per_nine_innings': None,
            'batters_faced': None,
            'batters_struckout_per_nine_innings': None,
            'earned_runs_allowed': None,
            'era': None,
            'era_plus': None,
            'fielding_independent_pitching': None,
            'games_finished': None,
            'hits_against_per_nine_innings': None,
            'hits_allowed': None,
            'home_runs_against_per_nine_innings': None,
            'home_runs_allowed': None,
            'intentional_bases_on_balls_given': None,
            'losses': None,
            'runs_allowed': None,
            'saves': None,
            'shutouts': None,
            'strikeouts': None,
            'strikeouts_thrown_per_walk': None,
            'times_hit_player': None,
            'whip': None,
            'wild_pitches': None,
            'win_percentage': None,
            'wins': None
        }

        self.player = Player('altuvjo01')

    def test_mlb_player_returns_requested_career_stats(self):
        # Request the career stats
        player = self.player('')

        for attribute, value in self.results_career.items():
            assert getattr(player, attribute) == value

    def test_mlb_player_returns_requested_player_season_stats(self):
        # Request the 2017 stats
        player = self.player('2017')

        for attribute, value in self.results_2017.items():
            assert getattr(player, attribute) == value

    def test_dataframe_returns_dataframe(self):
        dataframe = [
            {'assists': 2763,
             'at_bats': 4436,
             'bases_on_balls': 311,
             'batting_average': 0.317,
             'birth_date': '1990-05-06',
             'complete_games': 990,
             'defensive_chances': 4500,
             'defensive_runs_saved_above_average': -24,
             'defensive_runs_saved_above_average_per_innings': -3,
             'double_plays_turned': 623,
             'doubles': 271,
             'errors': 61,
             'fielding_percentage': 0.986,
             'games': 1105,
             'games_catcher': 0,
             'games_center_fielder': 0,
             'games_designated_hitter': 30,
             'games_first_baseman': 0,
             'games_in_batting_order': 1105,
             'games_in_defensive_lineup': 1070,
             'games_left_fielder': 0,
             'games_outfielder': 0,
             'games_pinch_hitter': 7,
             'games_pinch_runner': 4,
             'games_pitcher': 0,
             'games_right_fielder': 0,
             'games_second_baseman': 1070,
             'games_shortstop': 1,
             'games_started': 1059,
             'games_third_baseman': 0,
             'grounded_into_double_plays': 123,
             'height': '5-6',
             'hits': 1404,
             'home_runs': 96,
             'innings_played': 9269.0,
             'intentional_bases_on_balls': 38,
             'league_fielding_percentage': 0.984,
             'league_range_factor_per_game': 4.5,
             'league_range_factor_per_nine_innings': 4.53,
             'name': u'José Altuve',
             'nationality': 'Venezuela',
             'on_base_percentage': 0.364,
             'on_base_plus_slugging_percentage': 0.818,
             'on_base_plus_slugging_percentage_plus': 126,
             'plate_appearances': 4851,
             'player_id': 'altuvjo01',
             'position': '2B',
             'putouts': 1676,
             'range_factor_per_game': 4.15,
             'range_factor_per_nine_innings': 4.31,
             'runs': 635,
             'runs_batted_in': 457,
             'sacrifice_flies': 36,
             'sacrifice_hits': 24,
             'season': 'Career',
             'slugging_percentage': 0.454,
             'stolen_bases': 246,
             'team_abbreviation': 'HOU',
             'times_caught_stealing': 68,
             'times_hit_by_pitch': 44,
             'times_struck_out': 535,
             'total_bases': 2013,
             'total_fielding_runs_above_average': 23,
             'total_fielding_runs_above_average_per_innings': 3,
             'triples': 25,
             'weight': 165,
             # Properties specific to pitchers
             'balks': None,
             'bases_on_balls_given': None,
             'bases_on_balls_given_per_nine_innings': None,
             'batters_faced': None,
             'batters_struckout_per_nine_innings': None,
             'earned_runs_allowed': None,
             'era': None,
             'era_plus': None,
             'fielding_independent_pitching': None,
             'games_finished': None,
             'hits_against_per_nine_innings': None,
             'hits_allowed': None,
             'home_runs_against_per_nine_innings': None,
             'home_runs_allowed': None,
             'intentional_bases_on_balls_given': None,
             'losses': None,
             'runs_allowed': None,
             'saves': None,
             'shutouts': None,
             'strikeouts': None,
             'strikeouts_thrown_per_walk': None,
             'times_hit_player': None,
             'whip': None,
             'wild_pitches': None,
             'win_percentage': None,
             'wins': None},
            {'assists': 351,
             'at_bats': 590,
             'bases_on_balls': 58,
             'batting_average': 0.346,
             'birth_date': '1990-05-06',
             'complete_games': 134,
             'defensive_chances': 562,
             'defensive_runs_saved_above_average': 3,
             'defensive_runs_saved_above_average_per_innings': 3,
             'double_plays_turned': 86,
             'doubles': 39,
             'errors': 10,
             'fielding_percentage': 0.982,
             'games': 153,
             'games_catcher': 0,
             'games_center_fielder': 0,
             'games_designated_hitter': 3,
             'games_first_baseman': 0,
             'games_in_batting_order': 153,
             'games_in_defensive_lineup': 149,
             'games_left_fielder': 0,
             'games_outfielder': 0,
             'games_pinch_hitter': 1,
             'games_pinch_runner': 1,
             'games_pitcher': 0,
             'games_right_fielder': 0,
             'games_second_baseman': 149,
             'games_shortstop': 0,
             'games_started': 148,
             'games_third_baseman': 0,
             'grounded_into_double_plays': 19,
             'height': '5-6',
             'hits': 204,
             'home_runs': 24,
             'innings_played': 1283.1,
             'intentional_bases_on_balls': 3,
             'league_fielding_percentage': 0.983,
             'league_range_factor_per_game': 4.3,
             'league_range_factor_per_nine_innings': 4.35,
             'name': u'José Altuve',
             'nationality': 'Venezuela',
             'on_base_percentage': 0.41,
             'on_base_plus_slugging_percentage': 0.957,
             'on_base_plus_slugging_percentage_plus': 162,
             'plate_appearances': 662,
             'player_id': 'altuvjo01',
             'position': '2B',
             'putouts': 201,
             'range_factor_per_game': 3.7,
             'range_factor_per_nine_innings': 3.87,
             'runs': 112,
             'runs_batted_in': 81,
             'sacrifice_flies': 4,
             'sacrifice_hits': 1,
             'season': '2017',
             'slugging_percentage': 0.547,
             'stolen_bases': 32,
             'team_abbreviation': 'HOU',
             'times_caught_stealing': 6,
             'times_hit_by_pitch': 9,
             'times_struck_out': 84,
             'total_bases': 323,
             'total_fielding_runs_above_average': 13,
             'total_fielding_runs_above_average_per_innings': 12,
             'triples': 4,
             'weight': 165},
            {'assists': 361,
             'at_bats': 640,
             'bases_on_balls': 60,
             'batting_average': 0.338,
             'birth_date': '1990-05-06',
             'complete_games': 135,
             'defensive_chances': 574,
             'defensive_runs_saved_above_average': -2,
             'defensive_runs_saved_above_average_per_innings': -2,
             'double_plays_turned': 73,
             'doubles': 42,
             'errors': 7,
             'fielding_percentage': 0.988,
             'games': 161,
             'games_catcher': 0,
             'games_center_fielder': 0,
             'games_designated_hitter': 13,
             'games_first_baseman': 0,
             'games_in_batting_order': 161,
             'games_in_defensive_lineup': 148,
             'games_left_fielder': 0,
             'games_outfielder': 0,
             'games_pinch_hitter': 0,
             'games_pinch_runner': 1,
             'games_pitcher': 0,
             'games_right_fielder': 0,
             'games_second_baseman': 148,
             'games_shortstop': 1,
             'games_started': 147,
             'games_third_baseman': 0,
             'grounded_into_double_plays': 15,
             'height': '5-6',
             'hits': 216,
             'home_runs': 24,
             'innings_played': 1307.0,
             'intentional_bases_on_balls': 11,
             'league_fielding_percentage': 0.983,
             'league_range_factor_per_game': 4.44,
             'league_range_factor_per_nine_innings': 4.48,
             'name': u'José Altuve',
             'nationality': 'Venezuela',
             'on_base_percentage': 0.396,
             'on_base_plus_slugging_percentage': 0.928,
             'on_base_plus_slugging_percentage_plus': 155,
             'plate_appearances': 717,
             'player_id': 'altuvjo01',
             'position': '2B',
             'putouts': 206,
             'range_factor_per_game': 3.83,
             'range_factor_per_nine_innings': 3.9,
             'runs': 108,
             'runs_batted_in': 96,
             'sacrifice_flies': 7,
             'sacrifice_hits': 3,
             'season': '2016',
             'slugging_percentage': 0.531,
             'stolen_bases': 30,
             'team_abbreviation': 'HOU',
             'times_caught_stealing': 10,
             'times_hit_by_pitch': 7,
             'times_struck_out': 70,
             'total_bases': 340,
             'total_fielding_runs_above_average': 6,
             'total_fielding_runs_above_average_per_innings': 5,
             'triples': 5,
             'weight': 165,
             # Properties specific to pitchers
             'balks': None,
             'bases_on_balls_given': None,
             'bases_on_balls_given_per_nine_innings': None,
             'batters_faced': None,
             'batters_struckout_per_nine_innings': None,
             'earned_runs_allowed': None,
             'era': None,
             'era_plus': None,
             'fielding_independent_pitching': None,
             'games_finished': None,
             'hits_against_per_nine_innings': None,
             'hits_allowed': None,
             'home_runs_against_per_nine_innings': None,
             'home_runs_allowed': None,
             'intentional_bases_on_balls_given': None,
             'losses': None,
             'runs_allowed': None,
             'saves': None,
             'shutouts': None,
             'strikeouts': None,
             'strikeouts_thrown_per_walk': None,
             'times_hit_player': None,
             'whip': None,
             'wild_pitches': None,
             'win_percentage': None,
             'wins': None}
        ]
        indices = ['Career', '2017', '2016']

        df = pd.DataFrame(dataframe, index=indices)
        player = self.player('')

        # Pandas doesn't natively allow comparisons of DataFrames.
        # Concatenating the two DataFrames (the one generated during the test
        # and the expected one above) and dropping duplicate rows leaves only
        # the rows that are unique between the two frames. This allows a quick
        # check of the DataFrame to see if it is empty - if so, all rows are
        # duplicates, and they are equal.
        frames = [df, player.dataframe]
        df1 = pd.concat(frames).drop_duplicates(keep=False)

    def test_player_contract_returns_contract(self):
        contract = self.player.contract

        expected = {
            '2012': {
                'age': '22',
                'team': 'Houston Astros',
                'salary': '$483,000'
            },
            '2013': {
                'age': '23',
                'team': 'Houston Astros',
                'salary': '$505,700'
            },
            '2014': {
                'age': '24',
                'team': 'Houston Astros',
                'salary': '$1,250,000'
            },
            '2015': {
                'age': '25',
                'team': 'Houston Astros',
                'salary': '$2,500,000'
            },
            '2016': {
                'age': '26',
                'team': 'Houston Astros',
                'salary': '$3,500,000'
            },
            '2017': {
                'age': '27',
                'team': 'Houston Astros',
                'salary': '$4,500,000'
            },
            '2018': {
                'age': '28',
                'team': 'Houston Astros',
                'salary': '$9,000,000'
            },
            '2019': {
                'age': '29',
                'team': 'Houston Astros',
                'salary': '$9,500,000'
            },
            '2020': {
                'age': '30',
                'team': 'Houston Astros',
                'salary': '$29,000,000'
            },
            '2021': {
                'age': '31',
                'team': 'Houston Astros',
                'salary': '$29,000,000'
            },
            '2022': {
                'age': '32',
                'team': 'Houston Astros',
                'salary': '$29,000,000'
            },
            '2023': {
                'age': '33',
                'team': 'Houston Astros',
                'salary': '$29,000,000'
            },
            '2024': {
                'age': '34',
                'team': 'Houston Astros',
                'salary': '$29,000,000'
            }
            }

        assert contract == expected


class TestMLBPitcher:
    @mock.patch('requests.get', side_effect=mock_pyquery)
    def setup_method(self, *args, **kwargs):
        self.results_career = {
            'assists': 278,
            'at_bats': 48,
            'bases_on_balls': 0,
            'batting_average': 0.104,
            'birth_date': '1983-02-20',
            'complete_games': 24,
            'defensive_chances': 465,
            'defensive_runs_saved_above_average': 31,
            'defensive_runs_saved_above_average_per_innings': 2,
            'double_plays_turned': 28,
            'doubles': 0,
            'errors': 33,
            'fielding_percentage': 0.929,
            'games': 23,
            'games_catcher': 0,
            'games_center_fielder': 0,
            'games_designated_hitter': 0,
            'games_first_baseman': 0,
            'games_in_batting_order': 23,
            'games_in_defensive_lineup': 417,
            'games_left_fielder': 0,
            'games_outfielder': 0,
            'games_pinch_hitter': 0,
            'games_pinch_runner': 0,
            'games_pitcher': 417,
            'games_right_fielder': 0,
            'games_second_baseman': 0,
            'games_shortstop': 0,
            'games_started': 417,
            'games_third_baseman': 0,
            'grounded_into_double_plays': 1,
            'height': '6-5',
            'hits': 5,
            'home_runs': 0,
            'innings_played': 2747.0,
            'intentional_bases_on_balls': 0,
            'league_fielding_percentage': 0.953,
            'league_range_factor_per_game': 1.5,
            'league_range_factor_per_nine_innings': 1.51,
            'name': 'Justin Verlander',
            'nationality': 'United States of America',
            'on_base_percentage': 0.104,
            'on_base_plus_slugging_percentage': 0.208,
            'on_base_plus_slugging_percentage_plus': -43,
            'plate_appearances': 58,
            'player_id': 'verlaju01',
            'position': 'P',
            'putouts': 154,
            'range_factor_per_game': 1.04,
            'range_factor_per_nine_innings': 1.42,
            'runs': 2,
            'runs_batted_in': 1,
            'sacrifice_flies': 0,
            'sacrifice_hits': 10,
            'season': 'Career',
            'slugging_percentage': 0.104,
            'stolen_bases': 0,
            'team_abbreviation': 'HOU',
            'times_caught_stealing': 0,
            'times_hit_by_pitch': 0,
            'times_struck_out': 23,
            'total_bases': 5,
            'total_fielding_runs_above_average': None,
            'total_fielding_runs_above_average_per_innings': None,
            'triples': 0,
            'weight': 225,
            'balks': 19,
            'bases_on_balls_given': 807,
            'bases_on_balls_given_per_nine_innings': 2.6,
            'batters_faced': 11305,
            'batters_struckout_per_nine_innings': 8.8,
            'earned_runs_allowed': 1039,
            'era': 3.4,
            'era_plus': 125.0,
            'fielding_independent_pitching': 3.43,
            'games_finished': 0,
            'hits_against_per_nine_innings': 7.8,
            'hits_allowed': 2394,
            'home_runs_against_per_nine_innings': 0.9,
            'home_runs_allowed': 272,
            'intentional_bases_on_balls_given': 27,
            'losses': 123,
            'runs_allowed': 1124,
            'saves': 0,
            'shutouts': 8,
            'strikeouts': 2685,
            'strikeouts_thrown_per_walk': 3.33,
            'times_hit_player': 91,
            'whip': 1.165,
            'wild_pitches': 82,
            'win_percentage': 0.624,
            'wins': 204
        }

        self.results_2017 = {
            'assists': 13,
            'at_bats': 6,
            'bases_on_balls': 0,
            'batting_average': 0.167,
            'birth_date': '1983-02-20',
            'complete_games': 0,
            'defensive_chances': 24,
            'defensive_runs_saved_above_average': 0,
            'defensive_runs_saved_above_average_per_innings': 0,
            'double_plays_turned': 1,
            'doubles': 0,
            'errors': 1,
            'fielding_percentage': 0.958,
            'games': 2,
            'games_catcher': 0,
            'games_center_fielder': 0,
            'games_designated_hitter': 0,
            'games_first_baseman': 0,
            'games_in_batting_order': 2,
            'games_in_defensive_lineup': 28,
            'games_left_fielder': 0,
            'games_outfielder': 0,
            'games_pinch_hitter': 0,
            'games_pinch_runner': 0,
            'games_pitcher': 28,
            'games_right_fielder': 0,
            'games_second_baseman': 0,
            'games_shortstop': 0,
            'games_started': 33,
            'games_third_baseman': 0,
            'grounded_into_double_plays': 0,
            'height': '6-5',
            'hits': 1,
            'home_runs': 0,
            'innings_played': 206.0,
            'intentional_bases_on_balls': 0,
            'league_fielding_percentage': 0.948,
            'league_range_factor_per_game': 1.32,
            'league_range_factor_per_nine_innings': 1.33,
            'name': 'Justin Verlander',
            'nationality': 'United States of America',
            'on_base_percentage': 0.167,
            'on_base_plus_slugging_percentage': 0.333,
            'on_base_plus_slugging_percentage_plus': -10,
            'plate_appearances': 6,
            'player_id': 'verlaju01',
            'position': 'P',
            'putouts': 10,
            'range_factor_per_game': 0.7,
            'range_factor_per_nine_innings': 1.0,
            'runs': 0,
            'runs_batted_in': 1,
            'sacrifice_flies': 0,
            'sacrifice_hits': 0,
            'season': '2017',
            'slugging_percentage': 0.167,
            'stolen_bases': 0,
            'team_abbreviation': 'HOU',
            'times_caught_stealing': 0,
            'times_hit_by_pitch': 0,
            'times_struck_out': 4,
            'total_bases': 1,
            'total_fielding_runs_above_average': None,
            'total_fielding_runs_above_average_per_innings': None,
            'triples': 0,
            'weight': 225,
            'balks': 0,
            'bases_on_balls_given': 72,
            'bases_on_balls_given_per_nine_innings': 3.1,
            'batters_faced': 849,
            'batters_struckout_per_nine_innings': 9.6,
            'earned_runs_allowed': 77,
            'era': 3.36,
            'era_plus': 131.0,
            'fielding_independent_pitching': 3.84,
            'games_finished': 0,
            'hits_against_per_nine_innings': 7.4,
            'hits_allowed': 170,
            'home_runs_against_per_nine_innings': 1.2,
            'home_runs_allowed': 27,
            'intentional_bases_on_balls_given': 4,
            'losses': 8,
            'runs_allowed': 80,
            'saves': 0,
            'shutouts': 0,
            'strikeouts': 219,
            'strikeouts_thrown_per_walk': 3.04,
            'times_hit_player': 4,
            'whip': 1.175,
            'wild_pitches': 5,
            'win_percentage': 0.652,
            'wins': 15
        }

        self.player = Player('verlaju01')

    def test_mlb_player_returns_requested_career_stats(self):
        # Request the career stats
        player = self.player('')

        for attribute, value in self.results_career.items():
            assert getattr(player, attribute) == value

    def test_mlb_player_returns_requested_player_season_stats(self):
        # Request the 2017 stats
        player = self.player('2017')

        for attribute, value in self.results_2017.items():
            assert getattr(player, attribute) == value

    def test_dataframe_returns_dataframe(self):
        dataframe = [
            {'assists': 278,
             'at_bats': 48,
             'bases_on_balls': 0,
             'batting_average': 0.104,
             'birth_date': '1983-02-20',
             'complete_games': 24,
             'defensive_chances': 465,
             'defensive_runs_saved_above_average': 31,
             'defensive_runs_saved_above_average_per_innings': 2,
             'double_plays_turned': 28,
             'doubles': 0,
             'errors': 33,
             'fielding_percentage': 0.929,
             'games': 23,
             'games_catcher': 0,
             'games_center_fielder': 0,
             'games_designated_hitter': 0,
             'games_first_baseman': 0,
             'games_in_batting_order': 23,
             'games_in_defensive_lineup': 417,
             'games_left_fielder': 0,
             'games_outfielder': 0,
             'games_pinch_hitter': 0,
             'games_pinch_runner': 0,
             'games_pitcher': 417,
             'games_right_fielder': 0,
             'games_second_baseman': 0,
             'games_shortstop': 0,
             'games_started': 417,
             'games_third_baseman': 0,
             'grounded_into_double_plays': 1,
             'height': '6-5',
             'hits': 5,
             'home_runs': 0,
             'innings_played': 2747.0,
             'intentional_bases_on_balls': 0,
             'league_fielding_percentage': 0.953,
             'league_range_factor_per_game': 1.5,
             'league_range_factor_per_nine_innings': 1.51,
             'name': 'Justin Verlander',
             'nationality': 'United States of America',
             'on_base_percentage': 0.104,
             'on_base_plus_slugging_percentage': 0.208,
             'on_base_plus_slugging_percentage_plus': -43,
             'plate_appearances': 58,
             'player_id': 'verlaju01',
             'position': 'P',
             'putouts': 154,
             'range_factor_per_game': 1.04,
             'range_factor_per_nine_innings': 1.42,
             'runs': 2,
             'runs_batted_in': 1,
             'sacrifice_flies': 0,
             'sacrifice_hits': 10,
             'season': 'Career',
             'slugging_percentage': 0.104,
             'stolen_bases': 0,
             'team_abbreviation': 'HOU',
             'times_caught_stealing': 0,
             'times_hit_by_pitch': 0,
             'times_struck_out': 23,
             'total_bases': 5,
             'total_fielding_runs_above_average': None,
             'total_fielding_runs_above_average_per_innings': None,
             'triples': 0,
             'weight': 225,
             'balks': 19,
             'bases_on_balls_given': 807,
             'bases_on_balls_given_per_nine_innings': 2.6,
             'batters_faced': 11305,
             'batters_struckout_per_nine_innings': 8.8,
             'earned_runs_allowed': 1039,
             'era': 3.4,
             'era_plus': 125.0,
             'fielding_independent_pitching': 3.43,
             'games_finished': 0,
             'hits_against_per_nine_innings': 7.8,
             'hits_allowed': 2394,
             'home_runs_against_per_nine_innings': 0.9,
             'home_runs_allowed': 272,
             'intentional_bases_on_balls_given': 27,
             'losses': 123,
             'runs_allowed': 1124,
             'saves': 0,
             'shutouts': 8,
             'strikeouts': 2685,
             'strikeouts_thrown_per_walk': 3.33,
             'times_hit_player': 91,
             'whip': 1.165,
             'wild_pitches': 82,
             'win_percentage': 0.624,
             'wins': 204},
            {'assists': 13,
             'at_bats': 6,
             'bases_on_balls': 0,
             'batting_average': 0.167,
             'birth_date': '1983-02-20',
             'complete_games': 0,
             'defensive_chances': 24,
             'defensive_runs_saved_above_average': 0,
             'defensive_runs_saved_above_average_per_innings': 0,
             'double_plays_turned': 1,
             'doubles': 0,
             'errors': 1,
             'fielding_percentage': 0.958,
             'games': 2,
             'games_catcher': 0,
             'games_center_fielder': 0,
             'games_designated_hitter': 0,
             'games_first_baseman': 0,
             'games_in_batting_order': 2,
             'games_in_defensive_lineup': 28,
             'games_left_fielder': 0,
             'games_outfielder': 0,
             'games_pinch_hitter': 0,
             'games_pinch_runner': 0,
             'games_pitcher': 28,
             'games_right_fielder': 0,
             'games_second_baseman': 0,
             'games_shortstop': 0,
             'games_started': 33,
             'games_third_baseman': 0,
             'grounded_into_double_plays': 0,
             'height': '6-5',
             'hits': 1,
             'home_runs': 0,
             'innings_played': 206.0,
             'intentional_bases_on_balls': 0,
             'league_fielding_percentage': 0.948,
             'league_range_factor_per_game': 1.32,
             'league_range_factor_per_nine_innings': 1.33,
             'name': 'Justin Verlander',
             'nationality': 'United States of America',
             'on_base_percentage': 0.167,
             'on_base_plus_slugging_percentage': 0.333,
             'on_base_plus_slugging_percentage_plus': -10,
             'plate_appearances': 6,
             'player_id': 'verlaju01',
             'position': 'P',
             'putouts': 10,
             'range_factor_per_game': 0.7,
             'range_factor_per_nine_innings': 1.0,
             'runs': 0,
             'runs_batted_in': 1,
             'sacrifice_flies': 0,
             'sacrifice_hits': 0,
             'season': '2017',
             'slugging_percentage': 0.167,
             'stolen_bases': 0,
             'team_abbreviation': 'HOU',
             'times_caught_stealing': 0,
             'times_hit_by_pitch': 0,
             'times_struck_out': 4,
             'total_bases': 1,
             'total_fielding_runs_above_average': None,
             'total_fielding_runs_above_average_per_innings': None,
             'triples': 0,
             'weight': 225,
             'balks': 0,
             'bases_on_balls_given': 72,
             'bases_on_balls_given_per_nine_innings': 3.1,
             'batters_faced': 849,
             'batters_struckout_per_nine_innings': 9.6,
             'earned_runs_allowed': 77,
             'era': 3.36,
             'era_plus': 131.0,
             'fielding_independent_pitching': 3.84,
             'games_finished': 0,
             'hits_against_per_nine_innings': 7.4,
             'hits_allowed': 170,
             'home_runs_against_per_nine_innings': 1.2,
             'home_runs_allowed': 27,
             'intentional_bases_on_balls_given': 4,
             'losses': 8,
             'runs_allowed': 80,
             'saves': 0,
             'shutouts': 0,
             'strikeouts': 219,
             'strikeouts_thrown_per_walk': 3.04,
             'times_hit_player': 4,
             'whip': 1.175,
             'wild_pitches': 5,
             'win_percentage': 0.652,
             'wins': 15},
            {'assists': 29,
             'at_bats': 5,
             'bases_on_balls': 0,
             'batting_average': 0.2,
             'birth_date': '1983-02-20',
             'complete_games': 2,
             'defensive_chances': 37,
             'defensive_runs_saved_above_average': 5,
             'defensive_runs_saved_above_average_per_innings': 4,
             'double_plays_turned': 2,
             'doubles': 0,
             'errors': 1,
             'fielding_percentage': 0.973,
             'games': 2,
             'games_catcher': 0,
             'games_center_fielder': 0,
             'games_designated_hitter': 0,
             'games_first_baseman': 0,
             'games_in_batting_order': 2,
             'games_in_defensive_lineup': 34,
             'games_left_fielder': 0,
             'games_outfielder': 0,
             'games_pinch_hitter': 0,
             'games_pinch_runner': 0,
             'games_pitcher': 34,
             'games_right_fielder': 0,
             'games_second_baseman': 0,
             'games_shortstop': 0,
             'games_started': 34,
             'games_third_baseman': 0,
             'grounded_into_double_plays': 0,
             'height': '6-5',
             'hits': 1,
             'home_runs': 0,
             'innings_played': 227.2,
             'intentional_bases_on_balls': 0,
             'league_fielding_percentage': 0.95,
             'league_range_factor_per_game': 1.39,
             'league_range_factor_per_nine_innings': 1.41,
             'name': 'Justin Verlander',
             'nationality': 'United States of America',
             'on_base_percentage': 0.2,
             'on_base_plus_slugging_percentage': 0.4,
             'on_base_plus_slugging_percentage_plus': 9,
             'plate_appearances': 5,
             'player_id': 'verlaju01',
             'position': 'P',
             'putouts': 7,
             'range_factor_per_game': 1.06,
             'range_factor_per_nine_innings': 1.42,
             'runs': 1,
             'runs_batted_in': 0,
             'sacrifice_flies': 0,
             'sacrifice_hits': 0,
             'season': '2016',
             'slugging_percentage': 0.2,
             'stolen_bases': 0,
             'team_abbreviation': 'HOU',
             'times_caught_stealing': 0,
             'times_hit_by_pitch': 0,
             'times_struck_out': 2,
             'total_bases': 1,
             'total_fielding_runs_above_average': None,
             'total_fielding_runs_above_average_per_innings': None,
             'triples': 0,
             'weight': 225,
             'balks': 0,
             'bases_on_balls_given': 57,
             'bases_on_balls_given_per_nine_innings': 2.3,
             'batters_faced': 903,
             'batters_struckout_per_nine_innings': 10.0,
             'earned_runs_allowed': 77,
             'era': 3.04,
             'era_plus': 140.0,
             'fielding_independent_pitching': 3.48,
             'games_finished': 0,
             'hits_against_per_nine_innings': 6.8,
             'hits_allowed': 171,
             'home_runs_against_per_nine_innings': 1.2,
             'home_runs_allowed': 30,
             'intentional_bases_on_balls_given': 1,
             'losses': 9,
             'runs_allowed': 81,
             'saves': 0,
             'shutouts': 0,
             'strikeouts': 254,
             'strikeouts_thrown_per_walk': 4.46,
             'times_hit_player': 8,
             'whip': 1.001,
             'wild_pitches': 6,
             'win_percentage': 0.64,
             'wins': 16}
        ]
        indices = ['Career', '2017', '2016']

        df = pd.DataFrame(dataframe, index=indices)
        player = self.player('')

        # Pandas doesn't natively allow comparisons of DataFrames.
        # Concatenating the two DataFrames (the one generated during the test
        # and the expected one above) and dropping duplicate rows leaves only
        # the rows that are unique between the two frames. This allows a quick
        # check of the DataFrame to see if it is empty - if so, all rows are
        # duplicates, and they are equal.
        frames = [df, player.dataframe]
        df1 = pd.concat(frames).drop_duplicates(keep=False)

    def test_player_contract_returns_contract(self):
        contract = self.player.contract

        expected = {
            '2006': {
                'age': '23',
                'team': 'Detroit Tigers',
                'salary': '$980,000'
            },
            '2007': {
                'age': '24',
                'team': 'Detroit Tigers',
                'salary': '$1,030,000'
            },
            '2008': {
                'age': '25',
                'team': 'Detroit Tigers',
                'salary': '$1,130,000'
            },
            '2009': {
                'age': '26',
                'team': 'Detroit Tigers',
                'salary': '$3,675,000'
            },
            '2010': {
                'age': '27',
                'team': 'Detroit Tigers',
                'salary': '$6,850,000'
            },
            '2011': {
                'age': '28',
                'team': 'Detroit Tigers',
                'salary': '$12,850,000'
            },
            '2012': {
                'age': '29',
                'team': 'Detroit Tigers',
                'salary': '$20,000,000'
            },
            '2013': {
                'age': '30',
                'team': 'Detroit Tigers',
                'salary': '$20,000,000'
            },
            '2014': {
                'age': '31',
                'team': 'Detroit Tigers',
                'salary': '$20,000,000'
            },
            '2015': {
                'age': '32',
                'team': 'Detroit Tigers',
                'salary': '$28,000,000'
            },
            '2016': {
                'age': '33',
                'team': 'Detroit Tigers',
                'salary': '$28,000,000'
            },
            '2017': {
                'age': '34',
                'team': 'Houston Astros',
                'salary': '$28,000,000'
            },
            '2018': {
                'age': '35',
                'team': 'Houston Astros',
                'salary': '$28,000,000'
            },
            '2019': {
                'age': '36',
                'team': 'Houston Astros',
                'salary': '$28,000,000'
            }
            }

        assert contract == expected

    def test_correct_initial_index_found(self):
        seasons = ['2017', None, '2018']
        mock_season = mock.PropertyMock(return_value=seasons)
        player = Player('verlaju01')
        type(player)._season = mock_season

        result = player._find_initial_index()

        assert player._index == 1


class TestMLBRoster:
    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_roster_class_pulls_all_player_stats(self, *args, **kwargs):
        flexmock(utils) \
            .should_receive('_find_year_for_season') \
            .and_return('2017')
        roster = Roster('HOU')

        assert len(roster.players) == 3

        for player in roster.players:
            assert player.name in [u'José Altuve', 'Justin Verlander',
                                   'Charlie Morton']

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_bad_url_raises_value_error(self, *args, **kwargs):
        with pytest.raises(ValueError):
            roster = Roster('bad')

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_roster_from_team_class(self, *args, **kwargs):
        flexmock(Team) \
            .should_receive('_parse_team_data') \
            .and_return(None)
        team = Team(None, 1, '2018')
        mock_abbreviation = mock.PropertyMock(return_value='HOU')
        type(team)._abbreviation = mock_abbreviation

        assert len(team.roster.players) == 3

        for player in team.roster.players:
            assert player.name in [u'José Altuve', 'Justin Verlander',
                                   'Charlie Morton']
        type(team)._abbreviation = None

    @mock.patch('requests.get', side_effect=mock_pyquery)
    def test_roster_class_with_slim_parameter(self, *args, **kwargs):
        flexmock(utils) \
            .should_receive('_find_year_for_season') \
            .and_return('2018')
        roster = Roster('HOU', slim=True)

        assert len(roster.players) == 3
        assert roster.players == {
            'altuvjo01': 'Jose Altuve',
            'verlaju01': 'Justin Verlander',
            'mortoch02': 'Charlie Morton'
        }
